"""Handles conversion between the set of time intervals used in the `SosModel`

There are three main classes, which are currently rather intertwined.
:class:`Interval` represents an individual definition of a period
within a year.
This is specified using the ISO8601 period syntax and exposes
methods which use the isodate library to parse this into an internal hourly
representation of the period.

:class:`TimeIntervalRegister` holds the definitions of time-interval sets
specified for the sector models at the :class:`~smif.sos_model.SosModel`
level.
This class exposes one public method,
:py:meth:`~TimeIntervalRegister.add_interval_set` which allows the SosModel
to add an interval definition from a model configuration to the register.

Quantities
----------
Quantities are associated with a duration, period or interval.
For example 120 GWh of electricity generated during each week of February.::

        Week 1: 120 GW
        Week 2: 120 GW
        Week 3: 120 GW
        Week 4: 120 GW

Other examples of quantities:

- greenhouse gas emissions
- demands for infrastructure services
- materials use
- counts of cars past a junction
- costs of investments, operation and maintenance

Upscale: Divide
~~~~~~~~~~~~~~~

To convert to a higher temporal resolution, the values need to be apportioned
across the new time scale. In the above example, the 120 GWh of electricity
would be divided over the days of February to produce a daily time series
of generation.  For example::

        1st Feb: 17 GWh
        2nd Feb: 17 GWh
        3rd Feb: 17 GWh
        ...

Downscale: Sum
~~~~~~~~~~~~~~

To resample weekly values to a lower temporal resolution, the values
would need to be accumulated.  A monthly total would be::

        Feb: 480 GWh

Remapping
---------

Remapping quantities, as is required in the conversion from energy
demand (hourly values over a year) to energy supply (hourly values
for one week for each of four seasons) requires additional
averaging operations.  The quantities are averaged over the
many-to-one relationship of hours to time-slices, so that the
seasonal-hourly timeslices in the model approximate the hourly
profiles found across the particular seasons in the year. For example::

        hour 1: 20 GWh
        hour 2: 15 GWh
        hour 3: 10 GWh
        ...
        hour 8592: 16 GWh
        hour 8593: 12 GWh
        hour 8594: 21 GWh
        ...
        hour 8760: 43 GWh

To::

        season 1 hour 1: 20+16+.../4 GWh # Denominator number hours in sample
        season 1 hour 2: 15+12+.../4 GWh
        season 1 hour 3: 10+21+.../4 GWh
        ...

Prices
------

Unlike quantities, prices are associated with a point in time.
For example a spot price of £870/GWh.  An average price
can be associated with a duration, but even then,
we are just assigning a price to any point in time within a
range of times.

Upscale: Fill
~~~~~~~~~~~~~

Given a timeseries of monthly spot prices, converting these
to a daily price can be done by a fill operation.
E.g. copying the monthly price to each day.

From::

        Feb: £870/GWh

To::

        1st Feb: £870/GWh
        2nd Feb: £870/GWh
        ...

Downscale: Average
~~~~~~~~~~~~~~~~~~

On the other hand, going down scale, such as from daily prices
to a monthly price requires use of an averaging function. From::

        1st Feb: £870/GWh
        2nd Feb: £870/GWh
        ...

To::

        Feb: £870/GWh

Development Notes
-----------------

- We could use :py:meth:`numpy.convolve` to compare time intervals as hourly arrays
  before adding them to the set of intervals

"""
import logging
from datetime import datetime, timedelta

import numpy as np
from isodate import parse_duration
from smif.convert.register import NDimensionalRegister, ResolutionSet

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


"""Used as the reference year for computing time intervals
"""
BASE_YEAR = 2010


class Interval(object):
    """A time interval

    Parameters
    ----------
    id: str
        The unique name of the Interval
    list_of_intervals: str
        A list of tuples of valid ISO8601 duration definition
        string denoting the time elapsed from the beginning
        of the year to the (beginning, end) of the interval
    base_year: int, default=2010
        The reference year used for conversion to a datetime tuple

    Example
    -------

            >>> a = Interval('id', ('PT0H', 'PT1H'))
            >>> a.interval = ('PT1H', 'PT2H')
            >>> repr(a)
            "Interval('id', [('PT0H', 'PT1H'), ('PT1H', 'PT2H')], base_year=2010)"
            >>> str(a)
            "Interval 'id' starts at hour 0 and ends at hour 1"

    """
    def __init__(self, name, list_of_intervals, base_year=BASE_YEAR):
        self._name = name
        self._baseyear = base_year
        self.logger = logging.getLogger(__name__)

        if not list_of_intervals:
            msg = "Must construct Interval with at least one interval"
            raise ValueError(msg)

        if isinstance(list_of_intervals, list):
            for interval in list_of_intervals:
                assert isinstance(interval, tuple), "Interval must be constructed with tuples"
                if len(interval) != 2:
                    msg = "Interval tuple must take form (<start>, <end>)"
                    raise ValueError(msg)
            self._interval = list_of_intervals
        elif isinstance(list_of_intervals, tuple):
            self._interval = []
            self._interval.append(list_of_intervals)
        else:
            msg = "Interval tuple must take form (<start>, <end>)"
            raise ValueError(msg)

        self._validate()

    def _validate(self):
        for lower, upper in self.bounds:
            if lower > upper:
                msg = "A time interval must not end before it starts - found %d > %d"
                raise ValueError(msg, lower, upper)

    @property
    def name(self):
        """The name (or id) of the interval(s)

        Returns
        -------
        str
            Name or ID
        """
        return self._name

    @property
    def start(self):
        """The start hour of the interval(s)

        Returns
        -------
        list
            A list of integers, representing the hour from the beginning of the
            year associated with the start of each of the intervals
        """
        if len(self._interval) == 1:
            return self._interval[0][0]
        else:
            return [x[0] for x in self._interval]

    @property
    def end(self):
        """The end hour of the interval(s)

        Returns
        -------
        int or list
            An integer or list of integers, representing the hour from the
            beginning of the year associated with the end of each of the
            intervals
        """
        if len(self._interval) == 1:
            return self._interval[0][1]
        else:
            return [x[1] for x in self._interval]

    @property
    def interval(self):
        """The list of intervals

        Setter appends a tuple or list of intervals to the list of intervals
        """
        return sorted(self._interval)

    @interval.setter
    def interval(self, value):
        if isinstance(value, tuple):
            self._interval.append(value)
        elif isinstance(value, list):
            for element in value:
                assert isinstance(element, tuple), "A time interval must be a tuple"
            self._interval.extend(value)
        else:
            msg = "A time interval must add either a single tuple or a list of tuples"
            raise ValueError(msg)

        self._validate()

    @property
    def baseyear(self):
        """The reference year
        """
        return self._baseyear

    def __repr__(self):
        msg = "Interval('{}', {}, base_year={})"
        return msg.format(self.name, self.interval, self.baseyear)

    def __str__(self):
        string = "Interval '{}' maps to:\n".format(self.name)
        for interval in self.bounds:
            start = interval[0]
            end = interval[1]
            suffix = "  hour {} to hour {}\n".format(start, end)
            string += suffix

        return string

    def __eq__(self, other):
        if (self.name == other.name) \
           and (self.interval == other.interval) \
           and (self.baseyear == other.baseyear):
            return True
        else:
            return False

    @property
    def bounds(self):
        """Return a list of tuples of the intervals in terms of hours

        Returns
        -------
        list
            A list of tuples of the start and end hours of the year
            of the interval

        """
        hours = []
        for start_interval, end_interval in self.interval:
            start = self._convert_to_hours(start_interval)
            end = self._convert_to_hours(end_interval)
            hours.append((start, end))
        return hours

    def _convert_to_hours(self, duration):
        """

        Parameters
        ----------
        duration: str
            A valid ISO8601 duration definition string

        Returns
        -------
        int
            The hour in the year associated with the duration

        """
        reference = datetime(self._baseyear, 1, 1, 0)
        parsed_duration = parse_duration(duration)
        if isinstance(parsed_duration, timedelta):
            hours = parsed_duration.days * 24 + \
                    parsed_duration.seconds // 3600
        else:
            time = parsed_duration.totimedelta(reference)
            hours = time.days * 24 + time.seconds // 3600
        return hours

    def to_hourly_array(self):
        """Converts a list of intervals to a boolean array of hours

        Returns
        -------
        numpy.ndarray
            A boolean array
        """
        array = np.zeros(8760, dtype=np.int)
        for lower, upper in self.bounds:
            array[lower:upper] += 1
        return array


class IntervalSet(ResolutionSet):
    """A collection of intervals

    Arguments
    ---------
    name: str
        A unique identifier for the set of time intervals
    data: list
        Time intervals required as a list of dicts, with required keys
        ``start``, ``end`` and ``name``
    """

    def __init__(self, name, data, base_year=2010):
        self._data = []
        super().__init__()
        self.name = name
        self._base_year = base_year
        self.data = data
        self.bool_array = self._make_intersection_array()

    def _make_intersection_array(self):
        """
        Returns
        -------
        numpy.array
            A boolean array where rows correspond to entries in the interval
            set and columns represent hours of the year
        """
        array = np.zeros((len(self.data), 8760), dtype=np.bool)
        for row, interval in enumerate(self.data):
            array[row, :] = interval.to_hourly_array()
        return array

    @staticmethod
    def get_bounds(entry):
        return entry.bounds

    def get_proportion(self, from_index, to_interval):
        """Find proportion of interval address by `from_index` in `to_interval`

        Arguments
        ---------
        from_index : int
            Index of source interval
        to_interval : Interval
            The destination interval
        Returns
        -------
        float

        """
        from_interval = self.data[from_index]

        if len(from_interval.bounds) > 2:
            # Resampling
            proportion = self._compute_proportion(from_interval, to_interval)
            proportion = proportion * len(from_interval.bounds)

        elif len(to_interval.bounds) > 2:
            # Remapping
            proportion = self._compute_proportion(from_interval, to_interval)
            proportion = proportion / len(to_interval.bounds)

        else:
            proportion = self._compute_proportion(from_interval, to_interval)

        return proportion

    def _compute_proportion(self, from_interval, to_interval):
        from_hours = from_interval.to_hourly_array()
        to_hours = to_interval.to_hourly_array()

        # Find overlapping hours (intersection of two intervals)
        a_and_b = np.logical_and(from_hours, to_hours)
        # Find the proportion of from interval in the intersection of the
        # intervals
        intersection_duration = np.sum(a_and_b)
        from_duration = np.sum(from_hours)

        return intersection_duration / from_duration

    @property
    def coverage(self):
        """The total coverage in hours of the year by the interval set

        Returns
        -------
        float
        """
        compact_array = np.sum(self.bool_array, axis=1)
        coverage_value = np.sum(compact_array, axis=0)
        self.logger.debug("Coverage of %s is %s", self.name, coverage_value)
        return coverage_value

    def intersection(self, to_entry):
        """Return the destination intervals that intersect with `to_entry`

        Argument
        --------
        to_entry : Interval

        Returns
        -------
        list
            A list of Intervals that intersect with bounds

        Notes
        -----
        Look at the columns of the intersection array and identify
        overlapping intervals
        """
        elements = []

        for lower, upper in to_entry.bounds:
            bool_array = np.sum(self.bool_array[:, lower:upper], axis=1)
            intersect = np.nonzero(bool_array)[0]
            self.logger.debug(
                "Interval '%s' intersects with '%s'",
                to_entry.name, ",".join([self.data[x].name for x in intersect])
            )
            elements.extend(intersect)

        return elements

    @staticmethod
    def check_interval_bounds_equal(bounds):
        """Checks that each interval in the list of bounds is equal

        Arguments
        ---------
        bounds : list
            A list of tuples containing the start and end hours of an interval

        Returns
        -------
        bool
            True if all intervals in list of bounds are equal in length
        """
        return len(set([x[1] - x[0] for x in bounds])) <= 1

    @property
    def data(self):
        """Returns the intervals as a list

        Returns
        -------
        list
        """
        return self._data

    @data.setter
    def data(self, interval_data):
        """

        Arguments
        ---------
        interval_data : list
            A list of tuples containing (interval_id, list of interval tuples)
        """
        names = {}

        for interval in interval_data:
            name, interval_list = interval
            self._data.append(
                Interval(name, interval_list, self._base_year))
            names[name] = len(self._data) - 1

        self._validate_intervals()

    def _get_hourly_array(self):
        array = np.zeros(8760, dtype=np.int)
        for interval in self.data:
            array += interval.to_hourly_array()
        return array

    def _validate_intervals(self):
        if self.data:
            array = self._get_hourly_array()
            duplicate_hours = np.where(array > 1)[0]
            if len(duplicate_hours):
                hour = duplicate_hours[0]
                msg = "Duplicate entry for hour {} in interval set {}."
                raise ValueError(msg.format(hour, self.name))

    def get_entry_names(self):
        """Returns the names of the intervals
        """
        return [interval.name for interval in self.data]

    def __getitem___(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)


class TimeIntervalRegister(NDimensionalRegister):
    """Holds the set of time-intervals used by the SectorModels

    Notes
    -----
    The argument ``axis=1`` refers to the dimension of the data array that is
    associated with the intervals dimension.

    """
    def __init__(self):
        super().__init__(axis=1)


__REGISTER = TimeIntervalRegister()


def get_register():
    """Return single copy of TimeIntervalRegister
    """
    return __REGISTER



"""The Model Run collects scenarios, timesteps, narratives, and
model collection into a package which can be built and passed to
the ModelRunner to run.

The ModelRunner is responsible for running a ModelRun, including passing
in the correct data to the model between timesteps and calling to the
DecisionManager to obtain decisions.

ModeRun has attributes:
- id
- description
- sosmodel
- timesteps
- scenarios
- narratives
- strategy
- status

"""
from logging import getLogger

from smif.data_layer import DataHandle
from smif.decision import DecisionManager


class ModelRun(object):
    """Collects timesteps, scenarios , narratives and a SosModel together

    Attributes
    ----------
    name: str
        The unique name of the model run
    timestamp: :class:`datetime.datetime`
        An ISO8601 compatible timestamp of model run creation time
    description: str
        A friendly description of the model run
    sos_model: :class:`smif.model.sos_model.SosModel`
        The contained SosModel
    scenarios: dict
        For each scenario set, a mapping to a valid scenario within that set
    narratives: list
        A list of :class:`smif.parameters.Narrative` objects
    strategies: dict
    status: str
    logger: logging.Logger
    results: dict
    """

    def __init__(self):
        self.name = ""
        self.timestamp = None
        self.description = ""
        self.sos_model = None
        self._model_horizon = []

        self.scenarios = {}
        self.narratives = []
        self.strategies = None
        self.status = 'Empty'

        self.logger = getLogger(__name__)

        self.results = {}

    def as_dict(self):
        """Serialises :class:`smif.modelrun.ModelRun`

        Returns a dictionary definition of a ModelRun which is
        equivalent to that required by :class:`smif.modelrun.ModelRunBuilder`
        to construct a new model run

        Returns
        -------
        dict
        """
        config = {
            'name': self.name,
            'description': self.description,
            'stamp': self.timestamp,
            'timesteps': self._model_horizon,
            'sos_model': self.sos_model.name,
            'decision_module': None,
            'scenarios': self.scenarios,
            'narratives': self.narratives,
            'strategies': self.strategies
        }
        return config

    @property
    def model_horizon(self):
        """Returns the list of timesteps

        Returns
        =======
        list
            A list of timesteps, distinct and sorted in ascending order
        """
        return self._model_horizon.copy()

    @model_horizon.setter
    def model_horizon(self, value):
        self._model_horizon = sorted(list(set(value)))

    def run(self, store, warm_start_timestep=None):
        """Builds all the objects and passes them to the ModelRunner

        The idea is that this will add ModelRuns to a queue for asychronous
        processing

        """
        self.logger.debug("Running model run %s", self.name)
        if self.status == 'Built':
            if not self.model_horizon:
                raise ModelRunError("No timesteps specified for model run")
            if warm_start_timestep:
                idx = self.model_horizon.index(warm_start_timestep)
                self.model_horizon = self.model_horizon[idx:]
            self.status = 'Running'
            modelrunner = ModelRunner()
            modelrunner.solve_model(self, store)
            self.status = 'Successful'
        else:
            raise ModelRunError("Model is not yet built.")


class ModelRunner(object):
    """The ModelRunner orchestrates the simulation of a SoSModel over decision iterations and
    timesteps as provided by a DecisionManager.
    """
    def __init__(self):
        self.logger = getLogger(__name__)

    def solve_model(self, model_run, store):
        """Solve a ModelRun

        This method first calls :func:`smif.model.SosModel.before_model_run`
        with parameter data, then steps through the model horizon, calling
        :func:`smif.model.SosModel.simulate` with parameter data at each
        timestep.

        Arguments
        ---------
        model_run : :class:`smif.modelrun.ModelRun`
        """
        # Initialise each of the sector models
        self.logger.info("Initialising each of the sector models")
        data_handle = DataHandle(
            store=store,
            modelrun_name=model_run.name,
            current_timestep=None,
            timesteps=model_run.model_horizon,
            model=model_run.sos_model
        )
        model_run.sos_model.before_model_run(data_handle)

        # Initialise the decision manager (and hence decision modules)
        self.logger.debug("Initialising the decision manager")
        decision_manager = DecisionManager(model_run.model_horizon, model_run.strategies)

        # Solve the model run: decision loop generates a series of bundles of independent
        # decision iterations, each with a number of timesteps to run
        self.logger.debug("Solving the models over all timesteps: %s", model_run.model_horizon)
        for bundle in decision_manager.decision_loop():
            # each iteration is independent at this point, so the following loop is a
            # candidate for running in parallel
            for iteration, timesteps in bundle.items():
                self.logger.info('Running decision iteration %s', iteration)

                # Each timestep *might* be able to be run in parallel - until we have an
                # explicit way of declaring inter-timestep dependencies
                for timestep in timesteps:
                    self.logger.info('Running timestep %s', timestep)

                    # this should be pushed down into the responsibility of DataHandle and
                    # DataInterface
                    # - DecisionManager/Module sets state for each timestep, iteration
                    # - SosModel/Model calls through DataHandle to access 'current' state
                    state = decision_manager.get_state(timestep, iteration)

                    data_handle = DataHandle(
                        store=store,
                        modelrun_name=model_run.name,
                        current_timestep=timestep,
                        timesteps=model_run.model_horizon,
                        model=model_run.sos_model,
                        decision_iteration=iteration,
                        state=state
                    )
                    model_run.sos_model.simulate(data_handle)
        return data_handle


class ModelRunBuilder(object):
    """Builds the ModelRun object from the configuration
    """
    def __init__(self):
        self.model_run = ModelRun()
        self.logger = getLogger(__name__)

    def construct(self, model_run_config):
        """Set up the whole ModelRun

        Arguments
        ---------
        model_run_config : dict
            A valid model run configuration dictionary
        """
        self.model_run.name = model_run_config['name']
        self.model_run.description = model_run_config['description']
        self.model_run.timestamp = model_run_config['stamp']
        self._add_timesteps(model_run_config['timesteps'])
        self._add_sos_model(model_run_config['sos_model'])
        self._add_scenarios(model_run_config['scenarios'])
        self._add_narratives(model_run_config['narratives'])
        self._add_strategies(model_run_config['strategies'])

        self.model_run.status = 'Built'

    def finish(self):
        """Returns a configured model run ready for operation

        """
        if self.model_run.status == 'Built':
            return self.model_run
        else:
            raise RuntimeError("Run construct() method before finish().")

    def _add_sos_model(self, sos_model_object):
        """

        Arguments
        ---------
        sos_model_object : smif.model.sos_model.SosModel
        """
        self.model_run.sos_model = sos_model_object

    def _add_timesteps(self, timesteps):
        """Set the timesteps of the system-of-systems model

        Arguments
        ---------
        timesteps : list
            A list of timesteps
        """
        self.logger.info("Adding timesteps to model run")
        self.model_run.model_horizon = timesteps

    def _add_scenarios(self, scenarios):
        """

        Arguments
        ---------
        scenarios : dict
            A dictionary of {scenario set: scenario name}, one for each scenario set
        """
        self.model_run.scenarios = scenarios

    def _add_narratives(self, narratives):
        """

        Arguments
        ---------
        narratives : list
            A list of smif.parameters.Narrative objects
        """
        self.model_run.narratives = narratives

    def _add_strategies(self, strategies):
        """

        Arguments
        ---------
        narratives : list
            A list of strategies
        """
        self.model_run.strategies = strategies


class ModelRunError(Exception):
    """Raise when model run requirements are not satisfied
    """
    pass



"""Implements the interface to optimization functionality

This module defines an :class:`OptimizerTemplate` which lays out the fundamental
steps of any optimization algorithm. To implement your own algorithm, create a
child class which inherits OptimizerTemplate and implement the methods.

"""
__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


from abc import ABC, abstractmethod


class OptimizerTemplate(ABC):
    """
    """

    def __init__(self):
        self._action_space = None
        self._optimization_function = None
        self._results = {'optimal_decision': None,
                         'objective_value': None}

    def initialize(self, available_interventions):
        """Setup the optimization problem

        Arguments
        ---------
        available_interventions : int
            The number of dimensions
        """
        assert isinstance(available_interventions, int)
        self._action_space = available_interventions

    @property
    def optimization_function(self):
        """The function which will be minimized by the algorithm

        The optimization function should return a scalar and accept a vector
        of binary decision variables
        """
        return self._optimization_function

    @optimization_function.setter
    def optimization_function(self, function):
        self._optimization_function = function

    @property
    def results(self):
        """The results from a successful optimization
        """
        return self._results

    @results.setter
    def results(self, value):
        self._results = value

    @abstractmethod
    def run(self):
        """Override to implement an optimization algorithm

        The optimization algorithm should return the optimal decision vector
        and objective function
        """
        pass



import numpy as np
from smif.convert import SpaceTimeUnitConvertor


class TestSpaceTimeUnitConvertor_TimeOnly:

    def test_one_region_pass_through_time(self):
        """Only one region, 12 months, neither space nor time conversion is required
        """
        data = np.array([[
            31,
            28,
            31,
            30,
            31,
            30,
            31,
            31,
            30,
            31,
            30,
            31
        ]], dtype=float)
        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'half_squares',
            'months',
            'months',
            's',
            's'
        )
        assert np.allclose(actual, data)

    def test_one_region_time_aggregation(self):
        """Only one region, time aggregation is required
        """
        data = np.array([[
            31,
            28,
            31,
            30,
            31,
            30,
            31,
            31,
            30,
            31,
            30,
            31
        ]], dtype=float)  # area a, months 1-12

        expected = np.array([[
            31 + 31 + 28,
            31 + 30 + 31,
            30 + 31 + 31,
            30 + 31 + 30
        ]], dtype=float)  # area a, seasons 1-4

        convertor = SpaceTimeUnitConvertor()

        actual = convertor.convert(
            data,
            'half_squares',
            'half_squares',
            'months',
            'seasons',
            's',
            's'
        )
        assert np.allclose(actual, expected)

    def test_two_region_time_aggregation(self):
        """Two regions, time aggregation by region is required
        """
        data = np.array([
            # area a, months 1-12
            [
                31,
                28,
                31,
                30,
                31,
                30,
                31,
                31,
                30,
                31,
                30,
                31,
            ],
            # area b, months 1-12
            [
                31+1,
                28+1,
                31+1,
                30+1,
                31+1,
                30+1,
                31+1,
                31+1,
                30+1,
                31+1,
                30+1,
                31+1,
            ]
        ], dtype=float)

        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'half_squares',
            'months',
            'seasons',
            'm',
            'm'
        )

        expected = np.array([
            [
                31 + 31 + 28,
                31 + 30 + 31,
                30 + 31 + 31,
                30 + 31 + 30,
            ],
            [
                31 + 31 + 28 + 3,
                31 + 30 + 31 + 3,
                30 + 31 + 31 + 3,
                30 + 31 + 30 + 3,
            ]
        ], dtype=float)

        assert np.allclose(actual, expected)

    def test_one_region_convert_from_hour_to_day(self):
        """One region, time aggregation required
        """

        data = np.ones((1, 24))  # area a, hours 0-23
        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'half_squares',
            'hourly_day',
            'one_day',
            'm',
            'm'
        )
        expected = np.array([[24]])  # area a, day 0
        assert np.allclose(actual, expected)


class TestRemapConversion:

    def test_remap_timeslices_to_months(self):
        """One region, time remapping required
        """
        data = np.array([[
            1,  # winter month
            1,  # spring month
            1,  # summer month
            1  # autumn month
        ]], dtype=float)

        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'half_squares',
            'remap_months',
            'months',
            'm',
            'm'
        )
        expected = np.array([[1.03333333, 0.93333333, 1.01086957, 0.97826087,
                              1.01086957, 0.97826087, 1.01086957, 1.01086957,
                              0.98901099, 1.02197802, 0.98901099, 1.03333333]])
        assert np.allclose(actual, expected)

    def test_remap_months_to_timeslices(self, monthly_data, remap_month_data):
        """One region, time remapping required
        """
        data = monthly_data

        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'half_squares',
            'months',
            'remap_months',
            'm',
            'm'
        )
        expected = remap_month_data
        assert np.allclose(actual, expected)


class TestSpaceTimeUnitConvertor_RegionOnly:

    def test_half_to_one_region_pass_through_time(self):
        """Convert from half a region to one region, pass through time

        """

        data = np.ones((2, 12)) / 2  # area a,b, months 1-12

        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'rect',
            'months',
            'months',
            'm',
            'm'
        )
        expected = np.ones((1, 12))  # area zero, months 1-12
        assert np.allclose(actual, expected)

    def test_one_region_convert_from_hour_to_day(self):
        """Two regions aggregated to one, one interval
        """

        data = np.array([[24], [24]])  # area a,b, single interval

        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'rect',
            'one_day',
            'one_day',
            'm',
            'm'
        )
        expected = np.array([[48]])  # area zero, single interval
        assert np.allclose(actual, expected)


class TestSpaceTimeUnitConvertorBoth:

    def test_convert_both_space_and_time(self):
        data = np.ones((2, 12)) / 2  # area a,b, months 1-12

        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(
            data,
            'half_squares',
            'rect',
            'months',
            'seasons',
            'm',
            'm'
        )
        expected = np.ones((1, 4)) * 3  # area zero, seasons 1-4
        assert np.allclose(actual, expected)


class TestConvertor:

    def test_convertor(self):
        data = np.ones((2, 12)) / 2  # area a,b, months 1-12
        convertor = SpaceTimeUnitConvertor()
        actual = convertor.convert(data,
                                   'half_squares',
                                   'rect',
                                   'months',
                                   'seasons',
                                   'm',
                                   'm')
        expected = np.ones((1, 4)) * 3  # area zero, seasons 1-4
        assert np.allclose(actual, expected)



"""Tests parameter objects, passing and conversion into data dicts

- parameter objects
- that parameters are passed into the simulate method
"""
from pytest import fixture, raises
from smif.parameters import Parameter, ParameterList


@fixture
def config_list():
    return [
        {
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        },
        {
            'name': 'min_building_size',
            'description': 'Threshold for discarding polygons as not significant buildings',
            'absolute_range': (0, 1000000),
            'suggested_range': (0, 50),
            'default_value': 30,
            'units': 'm^2'
        }
    ]


@fixture
def parameters(config_list):
    return [
        Parameter.from_dict(config)
        for config in config_list
    ]


class TestParameter():
    """A `Parameter` holds metadata and provides creation/serialisation convenience methods
    """
    def test_default_creation(self):
        """should initialise with default values
        """
        parameter = Parameter()
        assert parameter.name == ''
        assert parameter.description == ''
        assert parameter.absolute_range == (0, 0)
        assert parameter.suggested_range == (0, 0)
        assert parameter.default_value == 0
        assert parameter.units == ''

    def test_creation(self):
        """should initialise with specified values
        """
        parameter = Parameter(
            name='test',
            description='Test description',
            absolute_range=(0, 1000),
            suggested_range=(0, 10),
            default_value=10,
            units='km'
        )
        assert parameter.name == 'test'
        assert parameter.description == 'Test description'
        assert parameter.absolute_range == (0, 1000)
        assert parameter.suggested_range == (0, 10)
        assert parameter.default_value == 10
        assert parameter.units == 'km'

    def test_dict_creation(self):
        """should initialise from dict
        """
        parameter = Parameter.from_dict({
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        })
        assert parameter.name == 'smart_meter_savings'
        assert parameter.description == 'The savings from smart meters'
        assert parameter.absolute_range == (0, 100)
        assert parameter.suggested_range == (3, 10)
        assert parameter.default_value == 3
        assert parameter.units == '%'

    def test_dict_creation_fail(self):
        """should fail if any field is missing from dict
        """
        parameter = Parameter.from_dict({
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'units': '%'
        })
        assert parameter.name == 'smart_meter_savings'
        assert parameter.description == 'The savings from smart meters'
        assert parameter.absolute_range == (0, 100)
        assert parameter.suggested_range == (3, 10)
        assert parameter.default_value == 3
        assert parameter.units == '%'

    def test_as_dict(self, config_list, parameters):
        """should serialise to plain dict
        """
        for expected, param in zip(config_list, parameters):
            actual = param.as_dict()
            assert actual == expected


class TestParameterList():
    """A `ParameterList` collects a set of parameter metadata, with convenience
    methods for deriving and overriding default values.
    """
    def test_default_creation(self):
        """should initialise empty
        """
        parameter_list = ParameterList()
        assert parameter_list.defaults == {}
        assert parameter_list.overridden({'test': 'value'}) == {}
        assert parameter_list.as_list() == []

    def test_creation_with_list_of_parameter(self, parameters):
        """should initialise with a list of `Parameter`, access dict-like by
        name
        """
        parameter_list = ParameterList(parameters)
        for parameter in parameters:
            assert parameter_list[parameter.name] == parameter

    def test_creation_with_list_of_dict(self, config_list, parameters):
        """should initialise with a list of dict suitable for `Parameter`
        creation
        """
        parameter_list = ParameterList(config_list)
        for parameter in parameters:
            assert parameter_list[parameter.name] == parameter

    def test_add_parameter(self, parameters):
        """should add a single parameter at a time
        """
        parameter = parameters[0]
        parameter_list = ParameterList()
        parameter_list.add_parameter(parameter)
        assert parameter_list[parameter.name] == parameter

    def test_add_dict(self, config_list, parameters):
        """should add a single dict at a time
        """
        config = config_list[0]
        parameter = parameters[0]
        parameter_list = ParameterList()
        parameter_list.add_parameter(config)
        assert parameter_list[parameter.name] == parameter

    def test_get_names(self, parameters):
        """should derive list of names
        """
        parameter_list = ParameterList(parameters)
        assert parameter_list.names == [parameter.name for parameter in parameters]

    def test_defaults(self, parameters):
        """should derive default values as dict of parameter_name => default_value
        """
        parameter_list = ParameterList(parameters)
        expected = {
            parameter.name: parameter.default_value
            for parameter in parameters
        }

        assert parameter_list.defaults == expected

    def test_overridden(self, parameters):
        """should override with configured values, without affecting defaults
        """
        parameter_list = ParameterList(parameters)

        defaults = {
            'smart_meter_savings': 3,
            'min_building_size': 30
        }
        override = {
            'smart_meter_savings': 5
        }
        expected = {
            'smart_meter_savings': 5,
            'min_building_size': 30
        }
        actual = parameter_list.overridden(override)

        assert parameter_list.defaults == defaults
        assert actual == expected

    def test_add_duplicate_parameter(self, parameters):
        """should fail to add the same parameter twice
        """
        parameter = parameters[0]

        parameter_list = ParameterList()
        parameter_list.add_parameter(parameter)
        with raises(ValueError):
            parameter_list.add_parameter(parameter)

    def test_as_list(self, config_list, parameters):
        """should return list of dicts suitable for serialisation or
        ParameterList creation
        """
        parameter_list = ParameterList(parameters)
        assert parameter_list.as_list() == config_list



from unittest.mock import Mock

import networkx
import numpy as np
import pytest
from pytest import fixture, raises
from smif.data_layer import DataHandle, MemoryInterface
from smif.metadata import Metadata, MetadataSet
from smif.model.scenario_model import ScenarioModel
from smif.model.sector_model import SectorModel as AbstractSectorModel
from smif.model.sos_model import ModelSet, SosModel


@fixture
def get_scenario():
    scenario = ScenarioModel('electricity_demand_scenario')
    scenario.scenario_name = 'Arbitrary Demand Scenario'
    scenario.add_output('electricity_demand_output',
                        scenario.regions.get_entry('LSOA'),
                        scenario.intervals.get_entry('annual'),
                        'unit')
    return scenario


@fixture
def get_sector_model():
    class SectorModel(AbstractSectorModel):
        """A no-op sector model
        """
        def simulate(self, timestep, data=None):
            return data

        def extract_obj(self):
            pass

        def initialise(self):
            pass
    return SectorModel


def get_data_handle(model):
    """Return a data handle for the model
    """
    store = MemoryInterface()
    store.write_sos_model_run({
        'name': 'test',
        'narratives': {}
    })
    store.write_scenario_data(
        'Arbitrary Demand Scenario',
        'electricity_demand_output',
        np.array([[123]]),
        'LSOA',
        'annual',
        2010)
    return DataHandle(
        store,
        'test',  # modelrun_name
        2010,  # current_timestep
        [2010],  # timesteps
        model
    )


@fixture
def get_water_sector_model():
    class SectorModel(AbstractSectorModel):
        """
        fluffyness -> electricity_demand
        """

        def simulate(self, data_handle=None):
            x = data_handle['fluffyness']
            data_handle['electricity_demand'] = (x**3) - (6 * x**2) + (0.9 * x) + 0.15
            return data_handle

        def extract_obj(self):
            pass

        def initialise(self):
            pass

    water_model = SectorModel('water_supply_model')
    water_model.add_input('fluffyness',
                          water_model.regions.get_entry('LSOA'),
                          water_model.intervals.get_entry('annual'),
                          'unit')
    water_model.add_output('electricity_demand',
                           water_model.regions.get_entry('LSOA'),
                           water_model.intervals.get_entry('annual'),
                           'unit')

    return water_model


@fixture
def get_energy_sector_model():
    class SectorModel(AbstractSectorModel):
        """
        electricity_demand_input -> fluffiness
        """

        def simulate(self, data_handle):
            """Mimics the running of a sector model
            """
            fluff = data_handle['electricity_demand_input']
            data_handle['fluffiness'] = fluff * 0.819
            return data_handle

        def extract_obj(self):
            pass

        def initialise(self):
            pass

    energy_model = SectorModel('energy_sector_model')
    energy_model.add_input('electricity_demand_input',
                           energy_model.regions.get_entry('LSOA'),
                           energy_model.intervals.get_entry('annual'),
                           'unit')
    energy_model.add_output('fluffiness',
                            energy_model.regions.get_entry('LSOA'),
                            energy_model.intervals.get_entry('annual'),
                            'unit')

    return energy_model


class TestModelSet:

    def test_model_set(self):
        elec_scenario = ScenarioModel('scenario')
        elec_scenario.add_output('output',
                                 elec_scenario.regions.get_entry('LSOA'),
                                 elec_scenario.intervals.get_entry('annual'),
                                 'unit')
        ModelSet({elec_scenario.name: elec_scenario})

    def test_model_set_deps(self, get_water_sector_model, get_energy_sector_model):
        pop_scenario = ScenarioModel('population')
        pop_scenario.add_output('population',
                                pop_scenario.regions.get_entry('LSOA'),
                                pop_scenario.intervals.get_entry('annual'),
                                'unit')
        energy_model = get_energy_sector_model
        energy_model.add_input('population',
                               energy_model.regions.get_entry('LSOA'),
                               energy_model.intervals.get_entry('annual'),
                               'unit')
        water_model = get_water_sector_model

        energy_model.add_dependency(pop_scenario, 'population', 'population')
        energy_model.add_dependency(water_model, 'electricity_demand',
                                    'electricity_demand_input')
        water_model.add_dependency(energy_model, 'fluffiness', 'fluffyness')

        model_set = ModelSet({
            energy_model.name: energy_model,
            water_model.name: water_model
        })
        # ModelSet should derive inputs as any input to one of its models which
        # is not met by an internal dependency
        print(model_set.inputs)
        assert len(model_set.inputs) == 1
        assert 'population' in model_set.inputs.names

        # ModelSet should derive dependencies as links to any model which
        # supplies a dependency not met internally
        assert len(model_set.deps) == 1
        assert model_set.deps['population'].source_model is pop_scenario


class TestBasics:

    def test_dependency_not_present(self, get_sector_model):
        SectorModel = get_sector_model
        elec_scenario = ScenarioModel('scenario')
        elec_scenario.add_output('output', Mock(), Mock(), 'unit')

        energy_model = SectorModel('model')
        energy_model.add_input('input', Mock(), Mock(), 'unit')
        with raises(ValueError):
            energy_model.add_dependency(elec_scenario, 'not_present',
                                        'input')

        with raises(ValueError):
            energy_model.add_dependency(elec_scenario, 'output',
                                        'not_correct_input_name')


class TestDependencyGraph:

    def test_simple_graph(self, get_sector_model):
        regions = Mock()
        regions.name = 'test_regions'
        intervals = Mock()
        intervals.name = 'test_intervals'

        SectorModel = get_sector_model
        elec_scenario = ScenarioModel('scenario')
        elec_scenario.add_output('output', regions, intervals, 'unit')

        energy_model = SectorModel('model')
        energy_model.add_input('input', regions, intervals, 'unit')
        energy_model.add_dependency(elec_scenario, 'output', 'input')

        sos_model = SosModel('energy_sos_model')
        sos_model.add_model(energy_model)
        sos_model.add_model(elec_scenario)

        # Builds the dependency graph
        sos_model.make_dependency_graph()

        graph = sos_model.dependency_graph

        assert energy_model in graph
        assert elec_scenario in graph

        assert list(graph.edges()) == [(elec_scenario, energy_model)]

    def test_get_model_sets(self, get_sector_model):
        regions = Mock()
        regions.name = 'test_regions'
        intervals = Mock()
        intervals.name = 'test_intervals'

        elec_scenario = ScenarioModel('scenario')
        elec_scenario.add_output('output', regions, intervals, 'unit')

        SectorModel = get_sector_model
        energy_model = SectorModel('model')
        energy_model.add_input('input', regions, intervals, 'unit')
        energy_model.add_dependency(elec_scenario, 'output', 'input')

        sos_model = SosModel('energy_sos_model')
        sos_model.add_model(energy_model)
        sos_model.add_model(elec_scenario)

        sos_model.make_dependency_graph()

        actual = sos_model._get_model_sets_in_run_order()
        expected = ['scenario', 'model']

        for model, name in zip(actual, expected):
            assert model.name == name

    def test_topological_sort(self, get_sector_model):
        regions = Mock()
        regions.name = 'test_regions'
        intervals = Mock()
        intervals.name = 'test_intervals'

        SectorModel = get_sector_model
        elec_scenario = ScenarioModel('scenario')
        elec_scenario.add_output('output', regions, intervals, 'unit')

        energy_model = SectorModel('model')
        energy_model.add_input('input', regions, intervals, 'unit')
        energy_model.add_dependency(elec_scenario, 'output', 'input')

        sos_model = SosModel('energy_sos_model')
        sos_model.add_model(energy_model)
        sos_model.add_model(elec_scenario)

        sos_model.make_dependency_graph()

        graph = sos_model.dependency_graph
        actual = list(networkx.topological_sort(graph))
        assert actual == [elec_scenario, energy_model]


class TestCompositeIntegration:

    def test_simplest_case(self, get_scenario):
        """One scenario only
        """
        elec_scenario = get_scenario
        sos_model = SosModel('simple')
        sos_model.add_model(elec_scenario)
        data_handle = get_data_handle(sos_model)
        sos_model.simulate(data_handle)
        # no results available, as only scenario model ran

    def test_sector_model_null_model(self, get_energy_sector_model):
        no_inputs = get_energy_sector_model
        no_inputs._inputs = MetadataSet([])
        no_inputs.simulate = lambda x: x

        data_handle = Mock()
        no_inputs.simulate(data_handle)
        data_handle.assert_not_called()

    def test_sector_model_one_input(self, get_energy_sector_model,
                                    get_scenario):
        elec_scenario = get_scenario

        energy_model = get_energy_sector_model

        sos_model = SosModel('blobby')
        sos_model.add_model(elec_scenario)
        sos_model.add_model(energy_model)
        energy_model.add_dependency(elec_scenario,
                                    'electricity_demand_output',
                                    'electricity_demand_input')

        data_handle = get_data_handle(sos_model)
        results = sos_model.simulate(data_handle)

        expected = np.array([[100.737]])
        actual = results.get_results('fluffiness', 'energy_sector_model')
        np.testing.assert_allclose(actual, expected, rtol=1e-5)


class TestNestedModels():

    def test_one_free_input(self, get_sector_model):
        SectorModel = get_sector_model
        energy_model = SectorModel('energy_sector_model')
        expected = Metadata('electricity_demand_input', Mock(), Mock(), 'unit')
        energy_model._inputs = MetadataSet([expected])

        actual = energy_model.free_inputs['electricity_demand_input']
        assert actual == expected

    def test_hanging_inputs(self, get_sector_model):
        """
        sos_model_high
            sos_model_lo
               -> em

        """
        SectorModel = get_sector_model
        energy_model = SectorModel('energy_sector_model')
        input_metadata = {
            'name': 'electricity_demand_input',
            'spatial_resolution': Mock(),
            'temporal_resolution': Mock(),
            'units': 'unit'
        }

        energy_model._inputs = MetadataSet([input_metadata])

        sos_model_lo = SosModel('lower')
        sos_model_lo.add_model(energy_model)

        expected = Metadata(input_metadata['name'],
                            input_metadata['spatial_resolution'],
                            input_metadata['temporal_resolution'],
                            input_metadata['units'])

        assert energy_model.free_inputs.names == ['electricity_demand_input']
        assert sos_model_lo.free_inputs.names == ['electricity_demand_input']

        sos_model_high = SosModel('higher')
        sos_model_high.add_model(sos_model_lo)
        actual = sos_model_high.free_inputs['electricity_demand_input']

        assert actual == expected

    @pytest.mark.xfail(reason="Nested sosmodels not yet implemented")
    def test_nested_graph(self, get_sector_model):
        """If we add a nested model, all Sectormodel and ScenarioModel objects
        are added as nodes in the graph with edges along dependencies.

        SosModel objects are not included, as they are just containers for the
        SectorModel and ScenarioModel objects, passing up inputs for deferred
        linkages to dependencies.

        Not implemented yet:
        """
        SectorModel = get_sector_model

        energy_model = SectorModel('energy_sector_model')

        input_metadata = {'name': 'electricity_demand_input',
                          'spatial_resolution': Mock(),
                          'temporal_resolution': Mock(),
                          'units': 'unit'}

        energy_model._inputs = MetadataSet([input_metadata])

        sos_model_lo = SosModel('lower')
        sos_model_lo.add_model(energy_model)

        sos_model_high = SosModel('higher')
        sos_model_high.add_model(sos_model_lo)

        with raises(NotImplementedError):
            sos_model_high.make_dependency_graph()
        graph = sos_model_high.dependency_graph
        assert graph.edges() == []

        expected = networkx.DiGraph()
        expected.add_node(sos_model_lo)
        expected.add_node(energy_model)

        assert energy_model in graph.nodes()

        scenario = ScenarioModel('electricity_demand')
        scenario.add_output('elec_demand_output', Mock(), Mock(), 'kWh')

        sos_model_high.add_dependency(scenario, 'elec_demand_output',
                                      'electricity_demand_input')

        sos_model_high.make_dependency_graph()
        assert graph.edges() == [(scenario, sos_model_high)]

    @pytest.mark.xfail(reason="Nested sosmodels not yet implemented")
    def test_composite_nested_sos_model(self, get_sector_model):
        """System of systems example with two nested SosModels, two Scenarios
        and one SectorModel. One dependency is defined at the SectorModel
        level, another at the lower SosModel level
        """
        SectorModel = get_sector_model

        elec_scenario = ScenarioModel('electricity_demand_scenario')
        elec_scenario.add_output('electricity_demand_output',
                                 Mock(), Mock(), 'unit')

        energy_model = SectorModel('energy_sector_model')
        energy_model.add_input(
            'electricity_demand_input', Mock(), Mock(), 'unit')
        energy_model.add_input('fluffiness_input', Mock(), Mock(), 'unit')
        energy_model.add_output('cost', Mock(), Mock(), 'unit')
        energy_model.add_output('fluffyness', Mock(), Mock(), 'unit')

        def energy_function(timestep, input_data):
            """Mimics the running of a sector model
            """
            results = {}
            demand = input_data['electricity_demand_input']
            fluff = input_data['fluffiness_input']
            results['cost'] = demand * 1.2894
            results['fluffyness'] = fluff * 22
            return results

        energy_model.simulate = energy_function
        energy_model.add_dependency(elec_scenario,
                                    'electricity_demand_output',
                                    'electricity_demand_input')

        sos_model_lo = SosModel('lower')
        sos_model_lo.add_model(elec_scenario)
        sos_model_lo.add_model(energy_model)

        fluf_scenario = ScenarioModel('fluffiness_scenario')
        fluf_scenario.add_output('fluffiness', Mock(), Mock(), 'unit')
        # fluf_scenario.add_data('fluffiness', np.array([[[12]]]), [2010])

        assert sos_model_lo.free_inputs.names == ['fluffiness_input']

        sos_model_lo.add_dependency(fluf_scenario,
                                    'fluffiness',
                                    'fluffiness_input')

        assert sos_model_lo.inputs.names == []

        sos_model_high = SosModel('higher')
        sos_model_high.add_model(sos_model_lo)
        sos_model_high.add_model(fluf_scenario)

        data_handle = get_data_handle(sos_model_high)
        actual = sos_model_high.simulate(data_handle)
        expected = {
            'fluffiness_scenario': {
                'fluffiness': 12
            },
            'lower': {
                'electricity_demand_scenario': {
                    'electricity_demand_output': 123
                },
                'energy_sector_model': {
                    'cost': 158.5962,
                    'fluffyness': 264
                }
            }
        }

        assert actual == expected


class TestCircularDependency:

    def test_loop(self, get_energy_sector_model, get_water_sector_model):
        """Fails because no functionality to deal with loops
        """
        energy_model = get_energy_sector_model
        water_model = get_water_sector_model

        sos_model = SosModel('energy_water_model')
        water_model.add_dependency(energy_model, 'fluffiness', 'fluffyness')
        energy_model.add_dependency(water_model, 'electricity_demand',
                                    'electricity_demand_input')
        sos_model.add_model(water_model)
        sos_model.add_model(energy_model)

        assert energy_model.inputs.names == ['electricity_demand_input']
        assert water_model.inputs.names == ['fluffyness']
        assert sos_model.inputs.names == []

        assert energy_model.free_inputs.names == []
        assert water_model.free_inputs.names == []
        assert sos_model.free_inputs.names == []

        sos_model.make_dependency_graph()
        graph = sos_model.dependency_graph

        assert (water_model, energy_model) in graph.edges()
        assert (energy_model, water_model) in graph.edges()

        sos_model.max_iterations = 100
        data_handle = get_data_handle(sos_model)
        results = sos_model.simulate(data_handle)

        expected = np.array([[0.13488114]], dtype=np.float)
        actual = results.get_results('fluffiness', model_name='energy_sector_model',
                                     modelset_iteration=35)
        np.testing.assert_allclose(actual, expected, rtol=1e-5)

        expected = np.array([[0.16469004]], dtype=np.float)
        actual = results.get_results('electricity_demand', model_name='water_supply_model',
                                     modelset_iteration=35)
        np.testing.assert_allclose(actual, expected, rtol=1e-5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""First approximation of a simulation models to be used in the framework

Simulation Models
-----------------

There are a bunch of very simple simulation models to demonstrate different
functionality.

`ExampleWaterSupplySimulation`: simulation only, no assets

`ExampleWaterSupplySimulationWithAsset`: simulation of water with outputs also
a function of assets

`ExampleWaterSupplySimulationWithReservoir`: simulation with outputs a function
of a reservoir level which should be persistent over years


Wrappers Around the Models
--------------------------

Around each of the above simulation models, subclassed wrappers based on
:class:`SectorModel` are used to present a consistent API

`WaterSupplyPython` - wraps `ExampleWaterSupplySimulation`

`WaterSupplyPythonAssets` - wraps `ExampleWaterSupplySimulationAsset`

`WaterSupplyExecutable` - wraps `water_supply_exec.py`

"""

import logging
import math
import numpy as np

from pytest import fixture
from smif.model.sector_model import SectorModel

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"

logger = logging.getLogger(__name__)


@fixture(scope='function')
def raininess_oracle(timestep):
    """Mimics an external data source for raininess

    Arguments
    =========
    timestep : int
        Requires a year between 2010 and 2050

    Returns
    =======
    raininess : int

    """
    msg = "timestep {} is outside of the range [2010, 2050]".format(timestep)
    assert timestep in [x for x in range(2010, 2051, 1)], msg

    raininess = math.floor((timestep - 2000) / 10)

    return raininess


def process_results(output):
    """Utility function which decodes stdout text from the water supply model

    Returns
    =======
    results : dict
        A dictionary where keys are the results e.g. `cost` and `water`

    """
    results = {}
    raw_results = output.decode('utf-8').split('\n')
    for result in raw_results[0:2]:
        values = result.split(',')
        if len(values) == 2:
            results[str(values[0])] = float(values[1])
    return results


class WaterSupplySectorModel(SectorModel):
    """Example of a class implementing the SectorModel interface,
    using one of the toy water models below to simulate the water supply
    system.
    """
    def initialise(self, initial_conditions):
        pass

    def simulate(self, timestep, data):
        """

        Parameters
        ----------
        decisions: list
        state: list
        data: dict
            A dict of parameters, as define by inputs.yaml, each entry of which
            contains a :class:`numpy.ndarray` with dimension regions x intervals

        """
        decisions = data[timestep]['decisions']

        # unpack inputs
        self.logger.debug(data)

        if 'raininess' in data[timestep]:
            raininess = data[timestep]['raininess'][0, 0]
        elif 'water' in data[timestep]:
            raininess = data[timestep]['water'][0, 0]
        else:
            raise KeyError("Couldn't find parameter in {}".format(data))

        # unpack decision variables
        if len(decisions) > 0:
            selective_list = [x for x in decisions
                              if x.name.startswith('water_asset')]
            number_of_treatment_plants = len(selective_list) + 1
        else:
            number_of_treatment_plants = 1

        # simulate (wrapping toy model)
        instance = ExampleWaterSupplySimulationModelWithAsset(
            raininess,
            number_of_treatment_plants
        )
        instance_results = instance.simulate(timestep, data)

        results = {
            'water': np.array([[
                instance_results['water']
            ]]),
            'cost': np.array([[
                instance_results['cost']
            ]]),
            'state': []
        }
        return {self.name: results}

    def extract_obj(self, results):
        return results['cost']

    def constraints(self, parameters):
        """

        Notes
        =====
        This constraint below expresses that water supply must be greater than
        or equal to 3.  ``x[0]`` is the decision variable for water treatment
        capacity, while the value ``parameters[0]`` in the min term is the
        value of the raininess parameter.
        """
        constraints = ({'type': 'ineq',
                        'fun': lambda x: min(x[0], parameters[0]) - 3})
        return constraints


class ExampleWaterSupplySimulationModel(object):
    """An example simulation model used for testing purposes

    Parameters
    ==========
    raininess : int
        The amount of rain produced in each simulation
    """
    def __init__(self, raininess):
        self.raininess = raininess
        self.water = None
        self.cost = None
        self.results = None

    def simulate(self):
        """Run the model

        Returns
        =======
        dict
        """
        self.water = self.raininess
        self.cost = 1
        return {
            "water": self.water,
            "cost": self.cost
        }


class ExampleWaterSupplySimulationModelWithAsset(ExampleWaterSupplySimulationModel):
    """An example simulation model which includes assets

    Parameters
    ==========
    raininess : int
        The amount of rain produced in each simulation
    number_of_treatment_plants : int
        The amount of water is a function of the number of treatment plants and
        the amount of raininess

    """
    def __init__(self, raininess, number_of_treatment_plants):
        """Overrides the basic example class to include treatment plants

        """
        self.number_of_treatment_plants = number_of_treatment_plants
        self.water = None
        self.cost = None
        super().__init__(raininess)

    def simulate(self, timestep, data=None):
        """Runs the water supply model

        Only 1 unit of water is produced per treatment plant,
        no matter how rainy.

        Each treatment plant costs 1.0 unit.
        """
        logger.debug("There are {} plants".format(
            self.number_of_treatment_plants))
        logger.debug("It is {} rainy".format(self.raininess))
        water = min(self.number_of_treatment_plants, self.raininess)
        cost = 1.264 * self.number_of_treatment_plants
        logger.debug("The system costs £{}".format(cost))
        return {
            "water": water,
            "cost": cost
        }


class ExampleWaterSupplySimulationModelWithReservoir(ExampleWaterSupplySimulationModel):
    """This simulation model has a state which is a non-asset variables

    The reservoir level is a function of the previous reservoir level,
    raininess and a fixed demand for water.

    Parameters
    ==========
    raininess : int
    reservoir_level : int

    Returns
    =======
    dict

    """
    fixed_demand = 1

    def __init__(self, raininess, reservoir_level):
        super().__init__(raininess)
        if reservoir_level < 0:
            raise ValueError("Reservoir level cannot be negative")
        self._reservoir_level = reservoir_level

    def simulate(self):
        """Run the model

        Note
        ====
        This simulate method has mixed the state transition (computing the new
        reservoir level) with the simulation of the model.

        """
        # Work out available water from raininess and initial reservoir level
        self.water = self.raininess + self._reservoir_level
        # Compute the reservoir level at the end of the year
        self._reservoir_level = self.water - self.fixed_demand
        self.cost = 1 + (0.1 * self._reservoir_level)
        return {
            'water': self.water,
            'cost': self.cost,
            'reservoir level': self._reservoir_level
        }



"""Tests to ensure that the example simulation model fixtures are behaving


"""
import os
import subprocess
import sys

from pytest import raises

from .fixtures.water_supply import (ExampleWaterSupplySimulationModel,
                                    ExampleWaterSupplySimulationModelWithReservoir,
                                    process_results, raininess_oracle)


def test_water_supply_with_reservoir():
    raininess = 1
    reservoir_level = 2
    model = ExampleWaterSupplySimulationModelWithReservoir(raininess, reservoir_level)

    actual = model.simulate()
    expected = {'cost': 1.2, 'water': 3, 'reservoir level': 2}
    assert actual == expected


def test_water_supply_with_reservoir_negative_level():
    raininess = 1
    reservoir_level = -2
    with raises(ValueError, message="Reservoir level cannot be negative"):
        ExampleWaterSupplySimulationModelWithReservoir(raininess, reservoir_level)


def test_process_results():
    input_bytes = b"cost,1\nwater,1\n"
    actual = process_results(input_bytes)
    expected = {'water': 1, 'cost': 1}
    assert actual == expected


def test_raininess_oracle():
    time = [2010, 2020, 2030, 2042, 2050]
    expected = [1, 2, 3, 4, 5]

    for result in zip(time, expected):
        actual = raininess_oracle(result[0])
        assert actual == result[1]


def test_raininess_oracle_out_of_range():
    msg = "timestep 2051 is outside of the range [2010, 2050]"
    with raises(AssertionError, message=msg):
        raininess_oracle(2051)


def test_simulate_rain_cost_python():
    raininess = 1
    model = ExampleWaterSupplySimulationModel(raininess)
    actual = model.simulate()
    expected = {'cost': 1, 'water': 1}
    assert actual == expected


def test_simulate_rain_executable():
    raininess = 10
    model_executable = sys.executable
    if model_executable != "" and model_executable is not None:
        model_script = os.path.join(os.path.dirname(__file__),
                                    "fixtures",
                                    "water_supply_exec.py")
        argument = "--raininess={}".format(str(raininess))
        output = subprocess.check_output([model_executable, model_script, argument])
        results = process_results(output)
        assert results['water'] == 10
        assert results['cost'] == 1



"""Contains classes and methods relating to narratives.

Narrative hold collections of overridden parameter data. During model setup,
a user compiles a number of narrative files which contains a list of parameter
names and values. These are assigned to a narrative set during a model run
and the Narrative object holds this information at runtime.
"""


class Narrative(object):
    """Holds information relating to parameters from a collection of narrative policies

    Arguments
    ---------
    name : str
    description :str
    narrative_set : str

    Example
    --------
    >>> narrative = Narrative('Energy Demand - High Tech',
                                  'A description',
                                  'technology')
    """
    def __init__(self, name, description, narrative_set):
        self._name = name
        self._description = description
        self._narrative_set = narrative_set

        self._data = {}

    @property
    def data(self):
        """The narrative data keyed by model name or ``global``

        Returns
        -------
        dict
            A nested dictionary containing the narrative data::

                {
                    'global': [
                        {'global_parameter': 'value'}
                    ],
                    'model_name': [
                        {'model_parameter': 'value'},
                        {'model_parameter_two': 'value'}
                    ]
                }
        """
        return self._data

    @data.setter
    def data(self, data):
        """Add data to the Narrative object

        Arguments
        ---------
        data : dict
            A dictionary of overridden parameter values

        Example
        -------
        >>> narrative_data = {
                'global': [
                    {'parameter_name': 42}
                ]
            }
        >>> narrative = Narrative(
                'Energy Demand - High Tech',
                'A description',
                'technology')
        >>> narrative.add_data(narrative_data)
        """
        if isinstance(data, dict):
            self._data.update(data)
        else:
            raise TypeError("Expected a dict of parameter values")

    def as_dict(self):
        """Serialise the narrative data

        Returns
        -------
        dict
            A dictionary of serialised narrative metadata::

                {'name': 'a_name',
                 'description': 'a description',
                 'narrative_set': 'a_narrative_set'}

        """
        config = {'name': self._name,
                  'description': self._description,
                  'narrative_set': self._narrative_set}
        return config

    def __eq__(self, other):
        return self.as_dict() == other.as_dict() and \
            self.data == other.data



"""Test HTTP API application
"""
import datetime
import json
import os
from unittest.mock import Mock

import pytest
import smif
from flask import current_app
from smif.data_layer import DataNotFoundError
from smif.http_api import create_app


@pytest.fixture
def mock_scheduler():

    def get_status(arg):
        if arg == 'model_never_started':
            return {
                'status': 'unstarted'
            }
        elif arg == 'model_started_and_running':
            return {
                'status': 'running',
            }
        elif arg == 'model_started_and_done':
            return {
                'status': 'done',
            }
        elif arg in ('unique_model_run_name'):
            return {
                'status': 'running',
            }

    attrs = {
        'get_status.side_effect': get_status
    }
    return Mock(**attrs)


@pytest.fixture
def mock_data_interface(get_sos_model_run, get_sos_model, get_sector_model,
                        get_scenario, get_scenario_set, get_narrative, get_narrative_set):

    def read_sos_model_run(arg):
        _check_exist('sos_model_run', arg)
        return get_sos_model_run

    def read_sos_model(arg):
        _check_exist('sos_model', arg)
        return get_sos_model

    def read_sector_model(arg):
        _check_exist('sector_model', arg)
        return get_sector_model

    def read_scenario(arg):
        _check_exist('scenario', arg)
        return get_scenario

    def read_scenario_set(arg):
        _check_exist('scenario_set', arg)
        return get_scenario_set

    def read_narrative(arg):
        _check_exist('narrative', arg)
        return get_narrative

    def read_narrative_set(arg):
        _check_exist('narrative_set', arg)
        return get_narrative_set

    def _check_exist(config, name):
        if name == 'does_not_exist':
            raise DataNotFoundError("%s '%s' not found" % (config, name))

    attrs = {
        'read_sos_model_runs.side_effect': [[get_sos_model_run]],
        'read_sos_model_run.side_effect':  read_sos_model_run,
        'read_sos_models.side_effect': [[get_sos_model]],
        'read_sos_model.side_effect':  read_sos_model,
        'read_sector_models.side_effect': [[get_sector_model]],
        'read_sector_model.side_effect':  read_sector_model,
        'read_scenarios.side_effect': [[get_scenario]],
        'read_scenario.side_effect':  read_scenario,
        'read_scenario_sets.side_effect': [[get_scenario_set]],
        'read_scenario_set.side_effect':  read_scenario_set,
        'read_narratives.side_effect': [[get_narrative]],
        'read_narrative.side_effect':  read_narrative,
        'read_narrative_sets.side_effect': [[get_narrative_set]],
        'read_narrative_set.side_effect':  read_narrative_set,
    }
    return Mock(**attrs)


@pytest.fixture
def app(request, mock_scheduler, mock_data_interface):

    """Return an app
    """
    test_app = create_app(
        static_folder=os.path.join(os.path.dirname(__file__), '..', 'fixtures', 'http'),
        template_folder=os.path.join(os.path.dirname(__file__), '..', 'fixtures', 'http'),
        data_interface=mock_data_interface,
        scheduler=mock_scheduler
    )

    with test_app.app_context():
        yield test_app


@pytest.fixture
def client(request, app):
    """Return an API client
    """
    test_client = app.test_client()

    def teardown():
        pass

    request.addfinalizer(teardown)
    return test_client


def parse_json(response):
    """Parse response data
    """
    return json.loads(response.data.decode('utf-8'))


def serialise_json(data):
    return json.dumps(data, default=timestamp_serialiser)


def timestamp_serialiser(obj):
    """Serialist datetime
    """
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()


def test_hello(client):
    """Start with a welcome message
    """
    response = client.get('/')
    assert "Welcome to smif" in str(response.data)


def test_get_smif(client):
    """GET smif details
    """
    response = client.get('/api/v1/smif/')
    data = parse_json(response)
    assert data['version'] == smif.__version__


def test_get_smif_version(client):
    """GET smif version
    """
    response = client.get('/api/v1/smif/version')
    data = parse_json(response)
    assert data == smif.__version__


def test_get_sos_model_runs(client, get_sos_model_run):
    """GET all model runs
    """
    response = client.get('/api/v1/sos_model_runs/')
    assert current_app.config.data_interface.read_sos_model_runs.call_count == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_sos_model_run]


def test_get_sos_model_runs_filtered_running(client, get_sos_model_run):
    """GET all model runs
    """
    response = client.get('/api/v1/sos_model_runs/?status=running')

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_sos_model_run]

def test_get_sos_model_run(client, get_sos_model_run):
    """GET single model run
    """
    name = get_sos_model_run['name']
    response = client.get('/api/v1/sos_model_runs/{}'.format(name))
    current_app.config.data_interface.read_sos_model_run.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_sos_model_run


def test_get_sos_model_run_missing(client):
    """GET missing system-of-systems model run
    """
    response = client.get('/api/v1/sos_model_runs/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "sos_model_run 'does_not_exist' not found"


def test_post_sos_model_run(client, get_sos_model_run):
    """POST model run
    """
    name = 'test_post_sos_model_run'
    get_sos_model_run['name'] = name
    send = serialise_json(get_sos_model_run)
    response = client.post(
        '/api/v1/sos_model_runs/',
        data=send,
        content_type='application/json')
    current_app.config.data_interface.write_sos_model_run.assert_called_with(get_sos_model_run)
    
    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_put_sos_model_run(client, get_sos_model_run):
    """PUT model run
    """
    send = serialise_json(get_sos_model_run)
    response = client.put(
        '/api/v1/sos_model_runs/' + get_sos_model_run['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_sos_model_run.assert_called_with(get_sos_model_run['name'], get_sos_model_run)

    assert response.status_code == 200


def test_delete_sos_model_run(client, get_sos_model_run):
    """DELETE sos_model_run
    """
    send = serialise_json(get_sos_model_run)
    response = client.delete(
        '/api/v1/sos_model_runs/' + get_sos_model_run['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_sos_model_run.assert_called_with(get_sos_model_run['name'])

    assert response.status_code == 200


def test_start_sos_model_run(client):
    """POST model run START
    """
    # Start a sos_model_run
    send = serialise_json({
            'args': {
                'verbosity': 0,
                'warm_start': False,
                'output_format': 'local_csv'
            }})    
    response = client.post(
        '/api/v1/sos_model_runs/20170918_energy_water/start',
        data=send,
        content_type='application/json')

    call = (current_app.config.scheduler.add.call_args)
    assert call[0][0] == '20170918_energy_water'
    assert call[0][1]['verbosity'] == 0
    assert call[0][1]['warm_start'] == False
    assert call[0][1]['output_format'] == 'local_csv'

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_kill_sos_model_run(client):
    """POST model run START
    """
    # Kill a sos_model_run
    response = client.post(
        '/api/v1/sos_model_runs/20170918_energy_water/kill',
        data={},
        content_type='application/json')
    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'

    current_app.config.scheduler.kill.assert_called_with('20170918_energy_water')


def test_get_modelrun_status_modelrun_never_started(client):
    """GET model run STATUS
    """
    # Check if the modelrun is running
    response = client.get(
        '/api/v1/sos_model_runs/model_never_started/status'
    )
    data = parse_json(response)
    assert response.status_code == 200
    assert data['status'] == 'unstarted'


def test_get_modelrun_status_modelrun_running(client):
    """GET model run STATUS
    """
    # Check if the modelrun is running
    response = client.get(
        '/api/v1/sos_model_runs/model_started_and_running/status'
    )
    data = parse_json(response)
    assert response.status_code == 200
    assert data['status'] == 'running'


def test_get_modelrun_status_modelrun_done(client):
    """GET model run STATUS
    """
    # Check if the modelrun was successful
    response = client.get(
        '/api/v1/sos_model_runs/model_started_and_done/status'
    )
    data = parse_json(response)
    assert response.status_code == 200
    assert data['status'] == 'done'


def test_get_sos_models(client, get_sos_model):
    """GET all system-of-systems models
    """
    response = client.get('/api/v1/sos_models/')
    assert current_app.config.data_interface.read_sos_models.called == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_sos_model]


def test_get_sos_model(client, get_sos_model):
    """GET single system-of-systems model
    """
    name = get_sos_model['name']
    response = client.get('/api/v1/sos_models/{}'.format(name))
    current_app.config.data_interface.read_sos_model.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_sos_model


def test_get_sos_model_missing(client):
    """GET missing system-of-systems model
    """
    response = client.get('/api/v1/sos_models/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "sos_model 'does_not_exist' not found"


def test_post_sos_model(client, get_sos_model):
    """POST system-of-systems model
    """
    name = 'test_post_sos_model'
    get_sos_model['name'] = name
    send = serialise_json(get_sos_model)
    response = client.post(
        '/api/v1/sos_models/',
        data=send,
        content_type='application/json')
    assert current_app.config.data_interface.write_sos_model.called == 1

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_put_sos_model(client, get_sos_model):
    """PUT sos_model
    """
    send = serialise_json(get_sos_model)
    response = client.put(
        '/api/v1/sos_models/' + get_sos_model['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_sos_model.assert_called_with(get_sos_model['name'], get_sos_model)
    
    assert response.status_code == 200


def test_delete_sos_model(client, get_sos_model):
    """DELETE sos_model
    """
    send = serialise_json(get_sos_model)
    response = client.delete(
        '/api/v1/sos_models/' + get_sos_model['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_sos_model.assert_called_with(get_sos_model['name'])

    assert response.status_code == 200


def test_get_sector_models(client, get_sector_model):
    """GET all model runs
    """
    response = client.get('/api/v1/sector_models/')
    assert current_app.config.data_interface.read_sector_models.called == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_sector_model]


def test_get_sector_model(client, get_sector_model):
    """GET single model run
    """
    name = get_sector_model['name']
    response = client.get('/api/v1/sector_models/{}'.format(name))
    current_app.config.data_interface.read_sector_model.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_sector_model


def test_get_sector_model_missing(client):
    """GET missing model run
    """
    response = client.get('/api/v1/sector_models/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "sector_model 'does_not_exist' not found"


def test_post_sector_model(client, get_sector_model):
    """POST sector model
    """
    name = 'test_post_sector_model'
    get_sector_model['name'] = name
    send = serialise_json(get_sector_model)
    response = client.post(
        '/api/v1/sector_models/',
        data=send,
        content_type='application/json')
    current_app.config.data_interface.write_sector_model.assert_called_with(get_sector_model)

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_put_sector_model(client, get_sector_model):
    """PUT sector_model
    """
    send = serialise_json(get_sector_model)
    response = client.put(
        '/api/v1/sector_models/' + get_sector_model['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_sector_model.assert_called_with(get_sector_model['name'], get_sector_model)
    
    assert response.status_code == 200


def test_delete_sector_model(client, get_sector_model):
    """DELETE sector_model
    """
    send = serialise_json(get_sector_model)
    response = client.delete(
        '/api/v1/sector_models/' + get_sector_model['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_sector_model.assert_called_with(get_sector_model['name'])

    assert response.status_code == 200


def test_get_scenario_sets(client, get_scenario_set):
    """GET all scenario_sets
    """
    response = client.get('/api/v1/scenario_sets/')
    assert current_app.config.data_interface.read_scenario_sets.called == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_scenario_set]


def test_get_scenario_set(client, get_scenario_set):
    """GET single system-of-systems model
    """
    name = get_scenario_set['name']
    response = client.get('/api/v1/scenario_sets/{}'.format(name))
    current_app.config.data_interface.read_scenario_set.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_scenario_set


def test_get_scenario_set_missing(client):
    """GET missing system-of-systems model
    """
    response = client.get('/api/v1/scenario_sets/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "scenario_set 'does_not_exist' not found"


def test_post_scenario_set(client, get_scenario_set):
    """POST system-of-systems model
    """
    name = 'test_post_scenario_set'
    get_scenario_set['name'] = name
    send = serialise_json(get_scenario_set)
    response = client.post(
        '/api/v1/scenario_sets/',
        data=send,
        content_type='application/json')
    current_app.config.data_interface.write_scenario_set.assert_called_with(get_scenario_set)

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_delete_scenario_set(client, get_scenario_set):
    """DELETE scenario_set
    """
    send = serialise_json(get_scenario_set)
    response = client.delete(
        '/api/v1/scenario_sets/' + get_scenario_set['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_scenario_set.assert_called_with(get_scenario_set['name'])

    assert response.status_code == 200


def test_put_scenario_set(client, get_scenario_set):
    """PUT scenario set
    """
    send = serialise_json(get_scenario_set)
    response = client.put(
        '/api/v1/scenario_sets/' + get_scenario_set['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_scenario_set.assert_called_with(get_scenario_set['name'], get_scenario_set)
    
    assert response.status_code == 200


def test_get_scenarios(client, get_scenario):
    """GET all scenarios
    """
    response = client.get('/api/v1/scenarios/')
    assert current_app.config.data_interface.read_scenarios.called == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_scenario]


def test_get_scenario(client, get_scenario):
    """GET single system-of-systems model
    """
    name = get_scenario['name']
    response = client.get('/api/v1/scenarios/{}'.format(name))
    current_app.config.data_interface.read_scenario.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_scenario


def test_get_scenario_missing(client):
    """GET missing system-of-systems model
    """
    response = client.get('/api/v1/scenarios/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "scenario 'does_not_exist' not found"


def test_post_scenario(client, get_scenario):
    """POST system-of-systems model
    """
    name = 'test_post_scenario'
    get_scenario['name'] = name
    send = serialise_json(get_scenario)
    response = client.post(
        '/api/v1/scenarios/',
        data=send,
        content_type='application/json')
    current_app.config.data_interface.write_scenario.assert_called_with(get_scenario)

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_delete_scenario(client, get_scenario):
    """DELETE scenario
    """
    send = serialise_json(get_scenario)
    response = client.delete(
        '/api/v1/scenarios/' + get_scenario['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_scenario.assert_called_with(get_scenario['name'])

    assert response.status_code == 200


def test_put_scenario(client, get_scenario):
    """PUT scenario
    """
    send = serialise_json(get_scenario)
    response = client.put(
        '/api/v1/scenarios/' + get_scenario['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_scenario.assert_called_with(get_scenario['name'], get_scenario)
    
    assert response.status_code == 200


def test_get_narrative_sets(client, get_narrative_set):
    """GET all narrative_sets
    """
    response = client.get('/api/v1/narrative_sets/')
    assert current_app.config.data_interface.read_narrative_sets.called == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_narrative_set]


def test_get_narrative_set(client, get_narrative_set):
    """GET single system-of-systems model
    """
    name = get_narrative_set['name']
    response = client.get('/api/v1/narrative_sets/{}'.format(name))
    current_app.config.data_interface.read_narrative_set.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_narrative_set


def test_get_narrative_set_missing(client):
    """GET missing system-of-systems model
    """
    response = client.get('/api/v1/narrative_sets/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "narrative_set 'does_not_exist' not found"


def test_post_narrative_set(client, get_narrative_set):
    """POST system-of-systems model
    """
    name = 'test_post_narrative_set'
    get_narrative_set['name'] = name
    send = serialise_json(get_narrative_set)
    response = client.post(
        '/api/v1/narrative_sets/',
        data=send,
        content_type='application/json')
    current_app.config.data_interface.write_narrative_set.assert_called_with(get_narrative_set)

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_put_narrative_set(client, get_narrative_set):
    """PUT narrative_set
    """
    send = serialise_json(get_narrative_set)
    response = client.put(
        '/api/v1/narrative_sets/' + get_narrative_set['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_narrative_set.assert_called_with(get_narrative_set['name'], get_narrative_set)
    
    assert response.status_code == 200


def test_delete_narrative_set(client, get_narrative_set):
    """DELETE narrative_set
    """
    send = serialise_json(get_narrative_set)
    response = client.delete(
        '/api/v1/narrative_sets/' + get_narrative_set['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_narrative_set.assert_called_with(get_narrative_set['name'])

    assert response.status_code == 200


def test_get_narratives(client, get_narrative):
    """GET all narratives
    """
    response = client.get('/api/v1/narratives/')
    assert current_app.config.data_interface.read_narratives.called == 1

    assert response.status_code == 200
    data = parse_json(response)
    assert data == [get_narrative]


def test_get_narrative(client, get_narrative):
    """GET single system-of-systems model
    """
    name = get_narrative['name']
    response = client.get('/api/v1/narratives/{}'.format(name))
    current_app.config.data_interface.read_narrative.assert_called_with(name)

    assert response.status_code == 200
    data = parse_json(response)
    assert data == get_narrative


def test_get_narrative_missing(client):
    """GET missing system-of-systems model
    """
    response = client.get('/api/v1/narratives/does_not_exist')
    assert response.status_code == 404
    data = parse_json(response)
    assert data['message'] == "narrative 'does_not_exist' not found"


def test_post_narrative(client, get_narrative):
    """POST system-of-systems model
    """
    name = 'test_post_narrative'
    get_narrative['name'] = name
    send = serialise_json(get_narrative)
    response = client.post(
        '/api/v1/narratives/',
        data=send,
        content_type='application/json')
    current_app.config.data_interface.write_narrative.assert_called_with(get_narrative)

    data = parse_json(response)
    assert response.status_code == 201
    assert data['message'] == 'success'


def test_put_narrative(client, get_narrative):
    """PUT narrative
    """
    send = serialise_json(get_narrative)
    response = client.put(
        '/api/v1/narratives/' + get_narrative['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.update_narrative.assert_called_with(get_narrative['name'], get_narrative)
    
    assert response.status_code == 200


def test_delete_narrative(client, get_narrative):
    """DELETE narrative
    """
    send = serialise_json(get_narrative)
    response = client.delete(
        '/api/v1/narratives/' + get_narrative['name'],
        data=send,
        content_type='application/json')
    current_app.config.data_interface.delete_narrative.assert_called_with(get_narrative['name'])

    assert response.status_code == 200


"""Wrap and solve a set of interdependent models

Given a directed graph of dependencies between models, any cyclic
dependencies are contained within the strongly-connected components of the
graph.

A ModelSet corresponds to the set of models within a single strongly-
connected component. This class provides the machinery necessary
to find a solution to each of the interdependent models.

The current implementation first estimates the outputs for each model in the
set, guaranteeing that each model will then be able to run, then begins
iterating, running every model in the set at each iteration, monitoring the
model outputs over the iterations, and stopping at timeout, divergence or
convergence.
"""

import numpy as np
from smif.data_layer import DataHandle
from smif.model import CompositeModel, element_before


class ModelSet(CompositeModel):
    """Wraps a set of interdependent models

    Parameters
    ----------
    models : dict
        A dict of model_name str => model :class:`smif.model.Model`
    max_iterations : int, default=25
        The maximum number of iterations that the model set will run before
        returning results
    relative_tolerance : float, default=1e-05
        Used to calculate when the model interations have converged
    absolute_tolerance : float, default=1e-08
        Used to calculate when the model interations have converged

    Attributes
    ----------
    max_iterations: int
        The maximum number of iterations
    models: list
        The list of :class:`smif.model.Model` subclasses
    """
    def __init__(self, models, max_iterations=25, relative_tolerance=1e-05,
                 absolute_tolerance=1e-08):
        name = "<->".join(sorted(model.name for model in models.values()))
        super().__init__(name)
        self.models = models
        self._model_names = list(models.keys())
        self._derive_deps_from_models()
        self._current_iteration = 0
        self._did_converge = False

        self.max_iterations = int(max_iterations)
        # tolerance for convergence assessment - see numpy.allclose docs
        self.relative_tolerance = float(relative_tolerance)
        self.absolute_tolerance = float(absolute_tolerance)

    def _derive_deps_from_models(self):
        for model in self.models.values():
            for sink, dep in model.deps.items():
                if dep.source_model not in self.models.values():
                    self.deps[sink] = dep
                    self.inputs.add_metadata(model.inputs[sink])

    def simulate(self, data_handle):
        """Runs a set of one or more models

        Arguments
        ---------
        timestep : int
        data : dict, default=None
        """
        self.logger.info("Simulating %s", self.name)

        # - keep track of intermediate results (iterations within the timestep)
        # - stop iterating according to near-equality condition
        for i in range(self.max_iterations):
            self._current_iteration = i
            if self._converged(data_handle):
                self._did_converge = True
                break
            else:
                self._run_iteration(i, data_handle)
        else:
            raise TimeoutError("Model evaluation exceeded max iterations")

    @property
    def max_iteration(self):
        """The maximum iteration reached before convergence
        """
        if self._did_converge:
            return self._current_iteration
        else:
            return None

    def _run_iteration(self, i, data_handle):
        """Run all models within the set

        Arguments
        ---------
        i : int
            Iteration counter
        data_handle : smif.data_layer.DataHandle
        """
        for model in self.models.values():
            self.logger.info("Simulating %s, iteration %s", model.name, i)
            model_data_handle = DataHandle(
                data_handle._store,
                data_handle._modelrun_name,
                data_handle._current_timestep,
                data_handle._timesteps,
                model,
                i,
                data_handle._decision_iteration
            )
            # Start by running all models in set with best guess
            # - zeroes
            # - last year's inputs
            if i == 0:
                self._guess_results(model, model_data_handle)
            else:
                model.simulate(model_data_handle)

    def _guess_results(self, model, data_handle):
        """Dependency-free guess at a model's input result set.

        Initially, guess zeroes, or the previous timestep's results.

        Arguments
        ---------
        model : smif.model.composite.Model
        data_handle : smif.data_layer.DataHandle

        Returns
        -------
        data_handle : smif.data_layer.DataHandle
        """
        timestep_before = element_before(
            data_handle.current_timestep,
            data_handle.timesteps
        )
        if timestep_before is not None:
            # last iteration of previous timestep results
            print(timestep_before)
            for output in model.outputs.metadata:
                data_handle.set_results(
                    output.name,
                    data_handle.get_results(output.name, timestep=timestep_before)
                )
        else:
            # generate zero-values for each parameter/region/interval combination
            for output in model.outputs.metadata:
                regions = output.get_region_names()
                intervals = output.get_interval_names()
                data_handle.set_results(
                    output.name,
                    np.zeros((len(regions), len(intervals)))
                )
        return data_handle

    def _converged(self, data_handle):
        """Check whether the results of a set of models have converged.

        Returns
        -------
        converged: bool
            True if the results have converged to within a tolerance

        Raises
        ------
        DiverganceError
            If the results appear to be diverging
        """
        if self._current_iteration < 2:
            # must have at least two result sets per model to assess convergence
            return False

        # each data output is a dict with
        #   str key (parameter name) =>
        #       np.ndarray value (regions x intervals)
        converged = []
        for model in self.models.values():
            model_data_handle = DataHandle(
                data_handle._store,
                data_handle._modelrun_name,
                data_handle._current_timestep,
                data_handle._timesteps,
                model,
                self._current_iteration,
                data_handle._decision_iteration
            )
            converged.append(self._model_converged(model, model_data_handle))

        if all(converged):
            # if all most recent are almost equal to penultimate, must have converged
            return True

        # TODO check for divergence and raise error

        return False

    def _model_converged(self, model, data_handle):
        """Check a single model's output for convergence

        Compare data output for each param over recent iterations.

        Parameters
        ----------
        model: :class:`smif.model.Model`
        data_handle: :class:`smif.data_layer.DataHandle`

        Returns
        -------
        bool
            True if converged otherwise, False
        """
        prev_data_handle = DataHandle(
            data_handle._store,
            data_handle._modelrun_name,
            data_handle._current_timestep,
            data_handle._timesteps,
            model,
            self._current_iteration - 1,  # access previous iteration
            data_handle._decision_iteration
        )
        return all(
            np.allclose(
                data_handle.get_data(param.name),
                prev_data_handle.get_data(param.name),
                rtol=self.relative_tolerance,
                atol=self.absolute_tolerance
            )
            for param in model.inputs.metadata
        )



from unittest.mock import Mock

from pytest import fixture
from smif.model.scenario_model import ScenarioModel, ScenarioModelBuilder


@fixture(scope='function')
def get_scenario_model_object():
    scenario_model = ScenarioModel('test_scenario_model')
    scenario_model.add_output('raininess',
                              scenario_model.regions.get_entry('LSOA'),
                              scenario_model.intervals.get_entry('annual'),
                              'ml')
    # data = np.array([[[3.]], [[5.]], [[1.]]], dtype=float)
    # scenario_model.add_data('raininess', data, [2010, 2011, 2012])
    return scenario_model


class TestScenarioObject:

    def test_serialise_scenario(self):
        scenario_model = ScenarioModel('High Population (ONS)')
        scenario_model.add_output('population_count',
                                  scenario_model.regions.get_entry('LSOA'),
                                  scenario_model.intervals.get_entry('annual'),
                                  'people')
        scenario_model.description = 'The High ONS Forecast for UK population out to 2050'
        scenario_model.scenario_set = 'population'
        actual = scenario_model.as_dict()
        expected = {
            'name': 'High Population (ONS)',
            'description': 'The High ONS Forecast for UK population out to 2050',
            'scenario_set': 'population',
            'facets': [
                {
                    'name': 'population_count',
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual',
                    'units': 'people'
                }
            ]
        }
        assert actual == expected

    def test_serialise_scenario_two_outputs(self, setup_folder_structure):
        scenario_model = ScenarioModel('High Population (ONS)')
        scenario_model.add_output('population_count',
                                  scenario_model.regions.get_entry('LSOA'),
                                  scenario_model.intervals.get_entry('annual'),
                                  'people')
        scenario_model.add_output('population_density',
                                  scenario_model.regions.get_entry('LSOA'),
                                  scenario_model.intervals.get_entry('annual'),
                                  'people / kilometer ** 2')
        scenario_model.description = 'The High ONS Forecast for UK population out to 2050'
        scenario_model.scenario_set = 'population'
        actual = scenario_model.as_dict()
        # sort to match expected output
        actual['facets'].sort(key=lambda p: p['name'])

        expected = {
            'name': 'High Population (ONS)',
            'description': 'The High ONS Forecast for UK population out to 2050',
            'scenario_set': 'population',
            'facets': [
                {
                    'name': 'population_count',
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual',
                    'units': 'people'
                },
                {
                    'name': 'population_density',
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual',
                    'units': 'people / kilometer ** 2'
                }
            ]
        }
        assert actual == expected


class TestScenarioModelData:

    def test_scenario_data(self):
        """Scenario model simulate method should be a no-op
        """
        builder = ScenarioModelBuilder('test_scenario_model')
        config = {
            'name': 'mass',
            'scenario_set': '',
            'facets': [
                {
                    'name': 'length',
                    'spatial_resolution': 'LSOA',
                    'temporal_resolution': 'annual',
                    'units': 'kg'
                }
            ]
        }
        builder.construct(config)
        scenario = builder.finish()

        data_handle = Mock()
        actual = scenario.simulate(data_handle)
        assert actual is data_handle
        data_handle.assert_not_called()



"""Test data interface
"""
import numpy as np
from pytest import raises
from smif.data_layer import DataMismatchError
from smif.data_layer.data_interface import DataInterface


class TestDataInterface():
    def test_data_list_to_array(self):

        data = [
            {
                'year': 2010,
                'value': 3,
                'region': 'oxford',
                'interval': '1'
            }
        ]
        actual = DataInterface.data_list_to_ndarray(
            data,
            ['oxford'],
            ['1']
        )
        expected = np.array([[3.]], dtype=float)
        np.testing.assert_equal(actual, expected)

    def test_scenario_data_missing_year(self, oxford_region):
        data = [
            {
                'value': 3.14
            }
        ]
        msg = "Observation missing region"
        with raises(KeyError) as ex:
            DataInterface.data_list_to_ndarray(
                data,
                ['oxford'],
                ['1']
            )
        assert msg in str(ex.value)

    def test_scenario_data_missing_param_region(self, oxford_region):
        data = [
            {
                'value': 3.14,
                'region': 'missing',
                'interval': '1',
                'year': 2015
            }
        ]
        msg = "Unknown region 'missing' in row 0"
        with raises(ValueError) as ex:
            DataInterface.data_list_to_ndarray(
                data,
                ['oxford'],
                ['1']
            )
        assert msg in str(ex)

    def test_scenario_data_missing_param_interval(self):
        data = [
            {
                'value': 3.14,
                'region': 'oxford',
                'interval': '1',
                'year': 2015
            },
            {
                'value': 3.14,
                'region': 'oxford',
                'interval': 'extra',
                'year': 2015
            }
        ]
        msg = "Number of observations (2) is not equal to intervals (1) x regions (1)"
        with raises(DataMismatchError) as ex:
            DataInterface.data_list_to_ndarray(
                data,
                ['oxford'],
                ['1']
            )
        assert msg in str(ex)



from unittest.mock import Mock

from pytest import fixture, raises
from smif.modelrun import ModelRunBuilder, ModelRunError, ModelRunner


@fixture(scope='function')
def get_model_run_config_data():

    sos_model = Mock()
    energy_supply = Mock()
    energy_supply.name = 'energy_supply'
    sos_model.models = [energy_supply]

    config = {
        'name': 'unique_model_run_name',
        'stamp': '2017-09-20T12:53:23+00:00',
        'description': 'a description of what the model run contains',
        'decision_module': None,
        'timesteps': [2010, 2011, 2012],
        'sos_model': sos_model,
        'scenarios':
            {'raininess': 'high_raininess'},
        'narratives':
            [Mock(data={'model_name': {'parameter_name': 0}}),
             Mock(data={'model_name': {'parameter_name': 0}})
             ],
        'strategies': [{'strategy': 'pre-specified-planning',
                        'description': 'build_nuclear',
                        'model_name': 'energy_supply',
                        'interventions': [
                            {'name': 'nuclear_large', 'build_year': 2012},
                            {'name': 'carrington_retire', 'build_year': 2011}]
                        }]
    }
    return config


@fixture(scope='function')
def get_model_run(get_model_run_config_data):

    config_data = get_model_run_config_data

    builder = ModelRunBuilder()
    builder.construct(config_data)
    return builder.finish()


class TestModelRunBuilder:

    def test_builder(self, get_model_run_config_data):

        config_data = get_model_run_config_data

        builder = ModelRunBuilder()
        builder.construct(config_data)

        modelrun = builder.finish()

        assert modelrun.name == 'unique_model_run_name'
        assert modelrun.timestamp == '2017-09-20T12:53:23+00:00'
        assert modelrun.model_horizon == [2010, 2011, 2012]
        assert modelrun.status == 'Built'
        assert modelrun.scenarios == {'raininess': 'high_raininess'}
        assert modelrun.narratives == config_data['narratives']
        assert modelrun.strategies == config_data['strategies']


class TestModelRun:

    def test_run_static(self, get_model_run):
        store = Mock()
        model_run = get_model_run
        model_run.run(store)

    def test_run_timesteps(self, get_model_run_config_data):
        """should error that timesteps are empty
        """
        config_data = get_model_run_config_data
        config_data['timesteps'] = []
        builder = ModelRunBuilder()
        builder.construct(config_data)
        model_run = builder.finish()
        store = Mock()
        with raises(ModelRunError) as ex:
            model_run.run(store)
        assert 'No timesteps specified' in str(ex)

    def test_serialize(self, get_model_run_config_data):
        builder = ModelRunBuilder()

        config = get_model_run_config_data

        builder.construct(config)
        model_run = builder.finish()

        config = model_run.as_dict()
        assert config == config


class TestModelRunner():

    def test_call_before_model_run(self):
        store = Mock()
        runner = ModelRunner()
        modelrun = Mock()
        modelrun.strategies = []
        sos_model = Mock()
        sos_model.models = []
        modelrun.sos_model = sos_model

        modelrun.narratives = []
        modelrun.model_horizon = [1, 2]

        runner.solve_model(modelrun, store)

        modelrun.sos_model.before_model_run.call_count == 1

    def test_call_simulate(self):
        store = Mock()
        runner = ModelRunner()
        modelrun = Mock()
        modelrun.strategies = []
        sos_model = Mock()
        sos_model.models = []
        modelrun.sos_model = sos_model
        modelrun.narratives = []
        modelrun.model_horizon = [1, 2]

        runner.solve_model(modelrun, store)

        assert modelrun.sos_model.simulate.call_count == 2



from smif import StateData


class TestStateData():
    """Test StateData object
    """
    def test_create(self):
        """StateData creation and accessors
        """
        val = StateData(1, {'test': "data"})

        assert val.target == 1
        assert val.data == {'test': "data"}



"""Implements procedures and tools to run a system-of-systems model
Provides methods for copying a sample project into a local directory
and for or running a system-of-systems model. Provides a Scheduler
that can run modelruns as a subprocess.

Example
-------
Copy a project folder
>>> copy_project_folder('/projects/smif/')

Run a single system-of-systems model:
>>> execute_model_run('energy_supply_demand', '/projects/smif/')

Use the Scheduler to run a system-of-systems model:
>>> my_scheduler = Scheduler()
>>> my_scheduler.add(
        'energy_supply_demand',
        {
            'directory': '/projects/smif'
        }
    )
>>> my_scheduler.status('energy_supply_demand')
{
    'message': 'running'
}
>>> my_scheduler.status('energy_supply_demand')
{
  "err": "",
  "output": "2010\n2010\n2015\n2015\nModel run '20170918_energy_water' complete\n",
  "status": "done"
}
"""

# import classes for access like ::
#         from smif.controller import Scheduler
from smif.controller.scheduler import Scheduler
from smif.controller.execute import execute_model_run
from smif.controller.setup import copy_project_folder

# Define what should be imported as * ::
#         from smif.controller import *
__all__ = ['Scheduler', 'execute_model_run', 'copy_project_folder']



"""HTTP API endpoint
"""
import dateutil.parser
import smif
from flask import current_app, jsonify, request
from flask.views import MethodView


class SmifAPI(MethodView):
    """Implement operations for Smif
    """
    def get(self, key):
        """Get smif details
        version: GET /api/v1/smif/version
        """
        if key == 'version':
            data = smif.__version__
        else:
            data = {}
            data['version'] = smif.__version__
        return jsonify(data)


class SosModelRunAPI(MethodView):
    """Implement CRUD operations for sos_model_run configuration data
    """
    def get(self, sos_model_run_name=None, action=None):
        """Get sos_model_runs
        all: GET /api/v1/sos_model_runs/
        one: GET /api/vi/sos_model_runs/name
        """
        data_interface = current_app.config.data_interface

        if action is None:
            if sos_model_run_name is None:

                sos_model_runs = data_interface.read_sos_model_runs()

                if 'status' in request.args.keys():
                    # filtered: GET /api/v1/sos_model_runs?status=done
                    data = []
                    for sos_model_run in sos_model_runs:
                        status = current_app.config.scheduler.get_status(sos_model_run['name'])
                        if status['status'] == request.args['status']:
                            data.append(sos_model_run)
                else:
                    # all: GET /api/v1/sos_model_runs/
                    data = sos_model_runs
            else:
                # one: GET /api/vi/sos_model_runs/name
                data = data_interface.read_sos_model_run(sos_model_run_name)
        elif action == 'status':
            # action: GET /api/vi/sos_model_runs/name/status
            data = current_app.config.scheduler.get_status(sos_model_run_name)

        return jsonify(data)

    def post(self, sos_model_run_name=None, action=None):
        """
        Create a sos_model_run:
        - POST /api/v1/sos_model_runs

        Perform an operation on a sos_model_run
        - POST /api/v1/sos_model_runs/<sos_model_run_name>/<action>

        Available actions are
        - start: Start the sos_model_run
        - kill: Stop a sos_model_run that is currently running
        - remove: Remove a sos_model_run that is waiting to be executed
        - resume: Warm start a sos_model_run
        """
        data_interface = current_app.config.data_interface

        if action is None:
            data = request.get_json() or request.form
            data_interface.write_sos_model_run(data)
        elif action == 'start':
            data = request.get_json() or request.form
            args = {
                'directory': data_interface.base_folder,
                'verbosity': data['args']['verbosity'],
                'warm_start': data['args']['warm_start'],
                'output_format': data['args']['output_format']
            }
            current_app.config.scheduler.add(sos_model_run_name, args)
        elif action == 'kill':
            current_app.config.scheduler.kill(sos_model_run_name)
        elif action == 'remove':
            raise NotImplementedError
        elif action == 'resume':
            raise NotImplementedError
        else:
            raise SyntaxError("SosModelRun action '%s' does not exist" % action)

        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, sos_model_run_name):
        """Update a sos_model_run:
        PUT /api/v1/sos_model_runs
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data_interface.update_sos_model_run(sos_model_run_name, data)
        response = jsonify({})
        return response

    def delete(self, sos_model_run_name):
        """Delete a sos_model_run:
        DELETE /api/v1/sos_model_runs
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_sos_model_run(sos_model_run_name)
        response = jsonify({})
        return response


class SosModelAPI(MethodView):
    """Implement CRUD operations for sos_model configuration data
    """
    def get(self, sos_model_name):
        """Get sos_model
        all: GET /api/v1/sos_model/
        one: GET /api/vi/sos_model/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface
        if sos_model_name is None:
            data = data_interface.read_sos_models()
            response = jsonify(data)
        else:
            data = data_interface.read_sos_model(sos_model_name)
            response = jsonify(data)

        return response

    def post(self):
        """Create a sos_model:
        POST /api/v1/sos_model
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form

        data_interface.write_sos_model(data)
        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, sos_model_name):
        """Update a sos_model:
        PUT /api/v1/sos_model
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data_interface.update_sos_model(sos_model_name, data)
        response = jsonify({})
        return response

    def delete(self, sos_model_name):
        """Delete a sos_model:
        DELETE /api/v1/sos_model
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_sos_model(sos_model_name)
        response = jsonify({})
        return response


class SectorModelAPI(MethodView):
    """Implement CRUD operations for sector_model configuration data
    """
    def get(self, sector_model_name):
        """Get sector_models
        all: GET /api/v1/sector_models/
        one: GET /api/vi/sector_models/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface
        if sector_model_name is None:
            data = data_interface.read_sector_models()
            response = jsonify(data)
        else:
            data = data_interface.read_sector_model(sector_model_name)
            response = jsonify(data)

        return response

    def post(self):
        """Create a sector_model:
        POST /api/v1/sector_models
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        data_interface.write_sector_model(data)
        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, sector_model_name):
        """Update a sector_model:
        PUT /api/v1/sector_models
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)
        data_interface.update_sector_model(sector_model_name, data)
        response = jsonify({})
        return response

    def delete(self, sector_model_name):
        """Delete a sector_model:
        DELETE /api/v1/sector_models
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_sector_model(sector_model_name)
        response = jsonify({})
        return response


class ScenarioSetAPI(MethodView):
    """Implement CRUD operations for scenario_sets configuration data
    """
    def get(self, scenario_set_name):
        """Get scenario_sets
        all: GET /api/v1/scenario_sets/
        one: GET /api/vi/scenario_sets/name
        """
        data_interface = current_app.config.data_interface
        if scenario_set_name is None:
            data = data_interface.read_scenario_sets()
            response = jsonify(data)
        else:
            data = data_interface.read_scenario_set(scenario_set_name)
            response = jsonify(data)

        return response

    def post(self):
        """Create a scenario_set:
        POST /api/v1/scenario_sets
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        data_interface.write_scenario_set(data)
        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, scenario_set_name):
        """Update a scenario_set:
        PUT /api/v1/scenario_sets
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)
        data_interface.update_scenario_set(scenario_set_name, data)
        response = jsonify({})
        return response

    def delete(self, scenario_set_name):
        """Delete a scenario_set:
        DELETE /api/v1/scenario_sets
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_scenario_set(scenario_set_name)
        response = jsonify({})
        return response


class ScenarioAPI(MethodView):
    """Implement CRUD operations for scenarios configuration data
    """
    def get(self, scenario_name):
        """Get scenarios
        all: GET /api/v1/scenarios/
        one: GET /api/vi/scenarios/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface
        if scenario_name is None:
            data = data_interface.read_scenarios()
            response = jsonify(data)
        else:
            data = data_interface.read_scenario(scenario_name)
            response = jsonify(data)

        return response

    def post(self):
        """Create a scenario:
        POST /api/v1/scenarios
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        data_interface.write_scenario(data)
        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, scenario_name):
        """Update a scenario:
        PUT /api/v1/scenarios
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)
        data_interface.update_scenario(scenario_name, data)
        response = jsonify({})
        return response

    def delete(self, scenario_name):
        """Delete a scenario:
        DELETE /api/v1/scenarios
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_scenario(scenario_name)
        response = jsonify({})
        return response


class NarrativeSetAPI(MethodView):
    """Implement CRUD operations for narrative_sets configuration data
    """
    def get(self, narrative_set_name):
        """Get narrative_sets
        all: GET /api/v1/narrative_sets/
        one: GET /api/vi/narrative_sets/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface
        if narrative_set_name is None:
            data = data_interface.read_narrative_sets()
            response = jsonify(data)
        else:
            data = data_interface.read_narrative_set(narrative_set_name)
            response = jsonify(data)

        return response

    def post(self):
        """Create a narrative_set:
        POST /api/v1/narrative_sets
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        data_interface.write_narrative_set(data)
        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, narrative_set_name):
        """Update a narrative_set:
        PUT /api/v1/narrative_sets
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)
        data_interface.update_narrative_set(narrative_set_name, data)
        response = jsonify({})
        return response

    def delete(self, narrative_set_name):
        """Delete a narrative_set:
        DELETE /api/v1/narrative_sets
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_narrative_set(narrative_set_name)
        response = jsonify({})
        return response


class NarrativeAPI(MethodView):
    """Implement CRUD operations for narratives configuration data
    """
    def get(self, narrative_name):
        """Get narratives
        all: GET /api/v1/narratives/
        one: GET /api/vi/narratives/name
        """
        # return str(current_app.config)
        data_interface = current_app.config.data_interface
        if narrative_name is None:
            data = data_interface.read_narratives()
            response = jsonify(data)
        else:
            data = data_interface.read_narrative(narrative_name)
            response = jsonify(data)

        return response

    def post(self):
        """Create a narrative:
        POST /api/v1/narratives
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)

        data_interface.write_narrative(data)
        response = jsonify({"message": "success"})
        response.status_code = 201
        return response

    def put(self, narrative_name):
        """Update a narrative:
        PUT /api/v1/narratives
        """
        data_interface = current_app.config.data_interface
        data = request.get_json() or request.form
        data = check_timestamp(data)
        data_interface.update_narrative(narrative_name, data)
        response = jsonify({})
        return response

    def delete(self, narrative_name):
        """Delete a narrative:
        DELETE /api/v1/narratives
        """
        data_interface = current_app.config.data_interface
        data_interface.delete_narrative(narrative_name)
        response = jsonify({})
        return response


def check_timestamp(data):
    """Check for timestamp and parse to datetime object
    """
    if 'stamp' in data:
        try:
            data['stamp'] = dateutil.parser.parse(data['stamp'])
        except(ValueError):
            pass
    return data



"""Common data interface
"""
from abc import ABCMeta, abstractmethod
from logging import getLogger

import numpy as np


class DataInterface(metaclass=ABCMeta):
    """Abstract base class to define common data interface
    """
    def __init__(self):
        self.logger = getLogger(__name__)

    @abstractmethod
    def read_coefficients(self, source_name, destination_name):
        raise NotImplementedError

    @abstractmethod
    def write_coefficients(self, source_name, destination_name, data):
        raise NotImplementedError

    @abstractmethod
    def read_units_file_name(self):
        raise NotImplementedError()

    @abstractmethod
    def read_sos_model_runs(self):
        raise NotImplementedError()

    @abstractmethod
    def read_sos_model_run(self, sos_model_run_name):
        raise NotImplementedError()

    @abstractmethod
    def write_sos_model_run(self, sos_model_run):
        raise NotImplementedError()

    @abstractmethod
    def update_sos_model_run(self, sos_model_run_name, sos_model_run):
        raise NotImplementedError()

    @abstractmethod
    def delete_sos_model_run(self, sos_model_run):
        raise NotImplementedError()

    @abstractmethod
    def read_sos_models(self):
        raise NotImplementedError()

    @abstractmethod
    def write_sos_model(self, sos_model):
        raise NotImplementedError()

    @abstractmethod
    def update_sos_model(self, sos_model_name, sos_model):
        raise NotImplementedError()

    @abstractmethod
    def read_sector_models(self):
        raise NotImplementedError()

    @abstractmethod
    def read_sector_model(self, sector_model_name):
        raise NotImplementedError()

    @abstractmethod
    def write_sector_model(self, sector_model):
        raise NotImplementedError()

    @abstractmethod
    def update_sector_model(self, sector_model_name, sector_model):
        raise NotImplementedError()

    @abstractmethod
    def read_region_definitions(self):
        raise NotImplementedError()

    @abstractmethod
    def read_region_definition_data(self, region_definition_name):
        raise NotImplementedError()

    @abstractmethod
    def write_region_definition(self, region_definition):
        raise NotImplementedError()

    @abstractmethod
    def update_region_definition(self, region_definition):
        raise NotImplementedError()

    @abstractmethod
    def read_interval_definitions(self):
        raise NotImplementedError()

    @abstractmethod
    def read_interval_definition_data(self, interval_definition_name):
        raise NotImplementedError()

    @abstractmethod
    def write_interval_definition(self, interval_definition):
        raise NotImplementedError()

    @abstractmethod
    def update_interval_definition(self, interval_definition):
        raise NotImplementedError()

    @abstractmethod
    def read_scenario_sets(self):
        raise NotImplementedError()

    @abstractmethod
    def read_scenario_set(self, scenario_set_name):
        raise NotImplementedError()

    @abstractmethod
    def read_strategies(self):
        raise NotImplementedError

    @abstractmethod
    def write_scenario_set(self, scenario_set):
        raise NotImplementedError()

    @abstractmethod
    def update_scenario_set(self, scenario_set):
        raise NotImplementedError()

    @abstractmethod
    def read_scenario_data(self, scenario_name, parameter_name,
                           spatial_resolution, temporal_resolution, timestep):
        raise NotImplementedError()

    @abstractmethod
    def write_scenario(self, scenario):
        raise NotImplementedError()

    @abstractmethod
    def update_scenario(self, scenario):
        raise NotImplementedError()

    @abstractmethod
    def read_narrative_sets(self):
        raise NotImplementedError()

    @abstractmethod
    def read_narrative_set(self, narrative_set_name):
        raise NotImplementedError()

    @abstractmethod
    def write_narrative_set(self, narrative_set):
        raise NotImplementedError()

    @abstractmethod
    def update_narrative_set(self, narrative_set):
        raise NotImplementedError()

    @abstractmethod
    def read_narrative_data(self, narrative_name):
        raise NotImplementedError()

    @abstractmethod
    def write_narrative(self, narrative):
        raise NotImplementedError()

    @abstractmethod
    def update_narrative(self, narrative):
        raise NotImplementedError()

    def read_parameters(self, modelrun_name, model_name):
        """Read global and model-specific parameter values for a given modelrun
        and model.
        """
        modelrun_config = self.read_sos_model_run(modelrun_name)
        params = {}
        for narratives in modelrun_config['narratives'].values():
            for narrative in narratives:
                data = self.read_narrative_data(narrative)
                for model_or_global, narrative_params in data.items():
                    if model_or_global in ('global', model_name):
                        params.update(narrative_params)
        return params

    @abstractmethod
    def read_results(self, modelrun_name, model_name, output_name, spatial_resolution,
                     temporal_resolution, timestep=None, modelset_iteration=None,
                     decision_iteration=None):
        raise NotImplementedError()

    @abstractmethod
    def write_results(self, modelrun_name, model_name, output_name, data, spatial_resolution,
                      temporal_resolution, timestep=None, modelset_iteration=None,
                      decision_iteration=None):
        raise NotImplementedError()

    @staticmethod
    def ndarray_to_data_list(data, region_names, interval_names, timestep=None):
        """Convert :class:`numpy.ndarray` to list of observations

        Parameters
        ----------
        data : numpy.ndarray
        region_names : list of str
        interval_names : list of str
        timestep: int or None

        Returns
        -------
        observations : list of dict
            Each dict has keys: 'region' (a region name), 'interval' (an
            interval name) and 'value'.
        """
        observations = []
        for region_idx, region in enumerate(region_names):
            for interval_idx, interval in enumerate(interval_names):
                observations.append({
                    'timestep': timestep,
                    'region': region,
                    'interval': interval,
                    'value': data[region_idx, interval_idx]
                })
        return observations

    @staticmethod
    def data_list_to_ndarray(observations, region_names, interval_names):
        """Convert list of observations to :class:`numpy.ndarray`

        Parameters
        ----------
        observations : list of dict
            Required keys for each dict are 'region' (a region name), 'interval'
            (an interval name) and 'value'.
        region_names : list
            A list of unique region names
        interval_names : list
            A list of unique interval names

        Returns
        -------
        data : numpy.ndarray

        Raises
        ------
        KeyError
            If an observation is missing a required key
        ValueError
            If an observation region or interval is not in region_names or
            interval_names
        DataNotFoundError
            If the observations don't include data for any region/interval
            combination
        """
        # Check that the list of region and interval names are unique
        assert len(region_names) == len(set(region_names))
        assert len(interval_names) == len(set(interval_names))

        DataInterface._validate_observations(observations, region_names, interval_names)

        data = np.full((len(region_names), len(interval_names)), np.nan)

        for obs in observations:
            region_idx = region_names.index(obs['region'])
            interval_idx = interval_names.index(obs['interval'])
            data[region_idx, interval_idx] = obs['value']

        return data

    @staticmethod
    def _validate_observations(observations, region_names, interval_names):
        if len(observations) != len(region_names) * len(interval_names):
            msg = "Number of observations ({}) is not equal to intervals ({}) x regions ({})"
            raise DataMismatchError(
                msg.format(len(observations), len(region_names), len(interval_names))
            )
        DataInterface._validate_observation_keys(observations)
        DataInterface._validate_observation_meta(observations, region_names, 'region')
        DataInterface._validate_observation_meta(observations, interval_names, 'interval')


    @staticmethod
    def _validate_observation_keys(observations):
        for obs in observations:
            if 'region' not in obs:
                raise KeyError(
                    "Observation missing region: {}".format(obs))
            if 'interval' not in obs:
                raise KeyError(
                    "Observation missing interval: {}".format(obs))
            if 'value' not in obs:
                raise KeyError(
                    "Observation missing value: {}".format(obs))

    @staticmethod
    def _validate_observation_meta(observations, meta_list, meta_name):
        observed = set()
        for line, obs in enumerate(observations):
            if obs[meta_name] not in meta_list:
                raise ValueError("Unknown {} '{}' in row {}".format(meta_name, obs[meta_name], line))
            else:
                observed.add(obs[meta_name])
        missing = set(meta_list) - observed
        if missing:
            raise DataNotFoundError("Missing values for %ss: %s", meta_name, list(missing))


class DataNotFoundError(Exception):
    """Raise when some data is not found
    """
    pass


class DataExistsError(Exception):
    """Raise when some data is found unexpectedly
    """
    pass


class DataMismatchError(Exception):
    """Raise when some data doesn't match the context

    E.g. when updating an object by id, the updated object's id must match
    the id provided separately.
    """
    pass



from logging import getLogger

from smif.convert import SpaceTimeUnitConvertor


class Dependency(object):
    """Link a model input to a data source

    Arguments
    ---------
    source_model : smif.model.Model
        The source model object
    source : smif.metadata.Metadata
        The source parameter (output) object
    sink : smif.metadata.Metadata
        The sink parameter (input) object
    function=None : func
        A conversion function
    """
    def __init__(self, source_model, source, sink, function=None):
        self.logger = getLogger(__name__)

        self.source_model = source_model
        self.source = source
        self.sink = sink

        # Insist on identical metadata - conversions to be explicit
        if source.spatial_resolution.name != \
                sink.spatial_resolution.name:
            self.logger.warn(
                "Implicit spatial conversion attempted ({}>{})".format(
                    source.spatial_resolution.name,
                    sink.spatial_resolution.name))
        if source.temporal_resolution.name != \
                sink.temporal_resolution.name:
            self.logger.warn(
                "Implicit spatial conversion attempted ({}>{})".format(
                    source.temporal_resolution.name,
                    sink.temporal_resolution.name))

        if function:
            self.convert = function
        else:
            self.convert = self._default_convert

    def _default_convert(self, data):
        """Convert dependency data

        Arguments
        ---------
        data : numpy.ndarray
            The data series for conversion
        """
        from_spatial = self.source.spatial_resolution.name
        to_spatial = self.sink.spatial_resolution.name
        from_temporal = self.source.temporal_resolution.name
        to_temporal = self.sink.temporal_resolution.name
        from_units = self.source.units
        to_units = self.sink.units

        if from_spatial != to_spatial \
                or from_temporal != to_temporal \
                or from_units != to_units:

            self.logger.debug("SpaceTimeUnit conversion: %s -> %s, %s -> %s, %s -> %s",
                              from_spatial, to_spatial, from_temporal, to_temporal,
                              from_units, to_units)
            convertor = SpaceTimeUnitConvertor()
            data = convertor.convert(data,
                                     from_spatial, to_spatial,
                                     from_temporal, to_temporal,
                                     from_units, to_units)

        return data

    def __repr__(self):
        return "<Dependency({}, {}, {})".format(self.source_model, self.source, self.sink)

    def __eq__(self, other):
        return self.source_model == other.source_model \
            and self.source == other.source \
            and self.sink == other.sink



import logging
import logging.config
import re
import sys

LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s %(name)-12s: %(levelname)-8s %(message)s'
        },
        'message': {
            'format': '\033[1;34m%(levelname)-8s\033[0m %(message)s'
        }
    },
    'handlers': {
        'file': {
            'class': 'logging.FileHandler',
            'level': 'DEBUG',
            'formatter': 'default',
            'filename': 'smif.log',
            'mode': 'a',
            'encoding': 'utf-8'
        },
        'stream': {
            'class': 'logging.StreamHandler',
            'formatter': 'message',
            'level': 'DEBUG'
        }
    },
    'root': {
        'handlers': ['file', 'stream'],
        'level': 'DEBUG'
    }
}

# Configure logging once, outside of any dependency on argparse
VERBOSITY = None
if '--verbose' in sys.argv:
    VERBOSITY = sys.argv.count('--verbose')
else:
    for arg in sys.argv:
        if re.match(r'\A-v+\Z', arg):
            VERBOSITY = len(arg) - 1
            break

if VERBOSITY is None:
    LOGGING_CONFIG['root']['level'] = logging.WARNING
elif VERBOSITY == 1:
    LOGGING_CONFIG['root']['level'] = logging.INFO
else:
    LOGGING_CONFIG['root']['level'] = logging.DEBUG

logging.config.dictConfig(LOGGING_CONFIG)
LOGGER = logging.getLogger(__name__)
LOGGER.debug('Debug logging enabled.')
